#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
class T(object):
    def __init__(self, *objs):
        self.__objs = list(objs)

    def add(self, *objs):
        self.__objs.extend(objs)

    def add_objects(self, objs):
        self.__objs.extend(objs)

    def iterate(self):
        return Iterator(self)

    def list(self):
        return self.__objs

    def __getitem__(self, idx):
        return self.__objs[idx]

    def nth(self, idx):
        return self.__objs[idx]


class Iterator(object):
    def __init__(self, set_):
        self.__set = set_
        self.__idx = 0

    def reset(self):
        self.__idx = 0

    def __next__(self):
        val = self.__set.nth(self.__idx)
        self.__idx += 1
        if self.__idx >= len(self.__set.list()):
            self.__idx = 0
        return val



# -*- coding: utf-8 -*-
# PIL is show-stoppers for this plugin
try:
    import PIL.Image
except:
    # image.py will complain about this one too, more verbosely
    print("""Python Imaging Library (PIL) not found, rendering of FITS files will not be available.
PIL can be installed from the Debian/Ubuntu package python-imaging, or can be downloaded from
http://www.pythonware.com/index.htm.
""")
    raise

try:
    from astropy.io import fits
except:
    print("""astropy package not found, rendering of FITS files will not be available.
""")
    raise

try:
    import numpy
    import scipy.ndimage.measurements
except:
    print("""numpy and/or scipy packages not found, rendering of FITS files will not be available.
Numpy and scipy can be installed from Debian/Ubuntu packages python-numpy and python-scipy, or can
be downloaded from http://numpy.scipy.org.
""")
    raise

# pychart needed for rendering histograms, but we can get on without it
from .local_pychart import *

pychart = True

# except:
#  pychart = None
#  print """PyChart package not found, rendering of FITS histograms will not be available.
# PyChart is available from http://home.gna.org/pychart/, or as Debian package python-pychart.
# """

import os.path
import traceback
import math
import pickle
import time

import Purr
from Purr.Render import quote_url, dprint, dprintf
from Purr.CachingRenderer import CachingRenderer
import Kittens.utils


class FITSRenderer(CachingRenderer):
    """This class renders FITS image data products."""

    @staticmethod
    def canRender(filename):
        """We can render it if PIL can read it."""
        if filename.endswith(".fits") or filename.endswith(".FITS"):
            return 10
        return False

    # this gives a short ID for the class (used in GUIs and such)
    renderer_id = "fits"

    # this gives a documentation string. You can use rich text here
    renderer_doc = """<P>The "fits" plugin provides rendering of FITS images."""

    # maximum thumbnail width & height
    # define renderer options
    CachingRenderer.addOption("image-thumbnail-width", 512, dtype=int, doc="Maximum width of thumbnails")
    CachingRenderer.addOption("image-thumbnail-height", 256, dtype=int, doc="Maximum height of thumbnails")
    CachingRenderer.addOption("hist-width", 1024, dtype=int, doc="Width of histogram plot")
    CachingRenderer.addOption("hist-height", 1024, dtype=int, doc="Height of histogram plot")
    CachingRenderer.addOption("hist-thumbnail-width", 128, dtype=int, doc="Maximum width of thumbnails")
    CachingRenderer.addOption("hist-thumbnail-height", 128, dtype=int, doc="Maximum height of thumbnails")
    CachingRenderer.addOption("fits-nimage", 4, dtype=int,
                              doc="Maximum number of planes to include, when dealing with cubes")
    CachingRenderer.addOption("fits-hist-nbin", 512, dtype=int, doc="Number of bins to use when making histograms""")
    CachingRenderer.addOption("fits-hist-clip", .95, dtype=float, doc="Apply histogram clipping")

    def _renderCacheFile(self, cachetype, relpath):
        return ("-cache-%s-rel-%s.html" % (cachetype, bool(relpath))).lower()

    @staticmethod
    def compute_tics(x0, x1):
        # work out tic interval
        dx = x1 - x0
        tic = 10 ** math.floor(math.log10(dx))
        if dx / tic < 2:
            tic /= 5
        elif dx / tic < 5:
            tic /= 2
        # work out where the tics are (in terms of i*interval)
        it0 = int(math.floor(x0 / tic))
        it1 = int(math.floor(x1 / tic))
        if it0 * tic < x0:
            it0 += 1
        # scale back up and return
        return [i * tic for i in range(it0, it1 + 1)]

    def _make_histogram(self, path, title, x, y):
        # make histogram by doubling up each x point to make "horizontals"
        x = list(map(float, x))
        y = list(map(int, y))
        width = x[1] - x[0]
        xy = []
        for a, b in zip(x, y):
            xy += [(a, b), (a + width, b)]
        # Make sure all spaces in path are escaped as pychart cannot handle them
        path = path.replace('\ ', ' ').replace(' ', '\ ')
        canv = canvas.init(path)
        ar = area.T(
            x_axis=axis.X(label="/20{}" + title, format="/20{}%g", tic_interval=self.compute_tics),
            y_axis=axis.Y(label=None, format="/20{}%s"),
            x_grid_style=line_style.gray70,
            y_grid_style=line_style.gray70,
            legend=None,
            size=self.hist_size
        )
        ar.add_plot(line_plot.T(
            data=xy, xcol=0, ycol=1,
            line_style=line_style.black,
            label=None,
            data_label_format=None
        ))
        ar.draw(canv)
        canv.close()

    def regenerate(self):
        Purr.progressMessage("reading %s" % self.dp.filename, sub=True)
        # init fitsfile to None, so that _read() above is forced to re-read it
        fitsfile = fits.open(self.dp.fullpath)
        header = fitsfile[0].header

        dprintf(3, "beginning render of", self.dp.fullpath);
        t0 = time.time()
        # write out FITS header
        self.headerfile, path, uptodate = self.subproduct("-fitsheader.html")
        if not uptodate:
            title = "FITS header for %s" % self.dp.filename
            html = """<HTML><BODY><TITLE>%s</TITLE>
      <H2>%s</H2>
      <PRE>""" % (title, title)
            for line in header.ascard:
                line = str(line).replace("<", "&lt;").replace(">", "&gt;")
                html += line + "\n"
            html += """
      </PRE></BODY></HTML>\n"""
            try:
                open(path, "w").write(html)
            except:
                print(("Error writing file %s" % path))
                traceback.print_exc()
                self.headerfile = None

        # figure out number of images to include
        ndim = header['NAXIS']
        fitsshape = [header['NAXIS%d' % i] for i in range(1, ndim + 1)]
        self.cubesize = 'x'.join(map(str, fitsshape))
        if ndim < 2:
            raise TypeError("can't render one-dimensional FITS files""")
        elif ndim == 2:
            fitsdata_to_images = lambda fdata: [fdata]
            nplanes = 1
        else:
            ax1 = ax2 = None
            # find the X/Y axes, by looking at CTYPEx
            # note that the array axes are in reverse order. I.e. if X is FITS axis 1 and Y is axis 2,
            # the array will be of e.g. shape 1,1,NY,NX, while fitsshape is [NX,NY,1,1]
            for i in range(1, ndim + 1):
                ctype = header['CTYPE%d' % i]
                if [prefix for prefix in ("RA", "GLON", "ELON", "HLON", "SLON") if ctype.startswith(prefix)] \
                        or ctype in ("L", "X"):
                    ax1 = ndim - i
                elif [prefix for prefix in ("DEC", "GLAT", "ELAT", "HLAT", "SLAT") if ctype.startswith(prefix)] \
                        or ctype in ("M", "Y"):
                    ax2 = ndim - i
            if ax1 is None or ax2 is None:
                ax1, ax2 = 1, 0
            arrshape = fitsshape[-1::-1]
            # this is how many planes we render, at most
            nplanes = max(self.getOption('fits-nimage'), 1)
            slices = []
            baseslice = [0] * ndim
            baseslice[ax1] = baseslice[ax2] = None
            imgshape = (arrshape[min(ax1, ax2)], arrshape[max(ax1, ax2)])
            while len(slices) < nplanes:
                slices.append(tuple(baseslice))
                for idim in range(ndim):
                    if baseslice[idim] != None:
                        baseslice[idim] += 1
                        if baseslice[idim] < arrshape[idim]:
                            break
                        else:
                            baseslice[idim] = 0
                else:
                    break
            nplanes = len(slices)

            # OK, slices contains how many slices to return
            def fitsdata_to_images(fdata, slices=slices, imgshape=imgshape):
                dprint(3, "fitsdata_to_images", slices, fdata.shape);
                t0 = time.time()
                # reshape to collapse into a 3D cube
                img = [fdata[i].reshape(imgshape) for i in slices]
                dprint(3, "collecting images took", time.time() - t0, "secs");
                t0 = time.time()
                return img

        # OK, now cycle over all images
        dprintf(3, "%s: rendering %d planes\n", self.dp.fullpath, nplanes);
        t0 = time.time()

        self.imgrec = [None] * nplanes
        # get number of bins (0 or None means no histogram)
        nbins = self.getOption("fits-hist-nbin")
        # see if histogram clipping is enabled, set hclip to None if not
        self.hclip = hclip = self.getOption("fits-hist-clip")
        if hclip == 1 or not nbins:
            hclip = None

        tsize_img = self.getOption("image-thumbnail-width"), self.getOption("image-thumbnail-height")
        tsize_hist = self.getOption("hist-thumbnail-width"), self.getOption("hist-thumbnail-height")
        self.hist_size = self.getOption("hist-width"), self.getOption("hist-height")

        # filled once we read the data
        images = None

        for num_image in range(nplanes):
            # do we have a cached status record for this image?
            recfile, recpath, uptodate = self.subproduct("-%d-stats" % num_image)
            if uptodate:
                dprintf(3, "%s(%d): stats file %s up-to-date, reading in\n", self.dp.fullpath, num_image, recfile)
                try:
                    self.imgrec[num_image] = pickle.load(open(recpath, "rb"))
                    continue
                except:
                    print(("Error reading stats file %s, regenerating everything" % recpath))
                    traceback.print_exc()
            # out of date, so we regenerate everything
            # build up record of stuff associated with this image
            rec = self.imgrec[num_image] = Kittens.utils.recdict()

            # generate paths for images
            rec.fullimage, img_path = self.subproductPath("-%d-full.png" % num_image)
            rec.thumbnail, img_thumb = self.subproductPath("-%d-thumb.png" % num_image)
            if pychart:
                rec.histogram_full, hf_path = self.subproductPath("-%d-hist-full.png" % num_image)
                rec.histogram_zoom, hz_path = self.subproductPath("-%d-hist-zoom.png" % num_image)
                rec.histogram_full_thumb, hf_thumb = self.subproductPath("-%d-hist-full-thumb.png" % num_image)
                rec.histogram_zoom_thumb, hz_thumb = self.subproductPath("-%d-hist-zoom-thumb.png" % num_image)

            # need to read in data at last
            if not images:
                dprint(3, "reading data");
                t0 = time.time()
                fitsdata = fitsfile[0].data
                dprint(3, "reading data took", time.time() - t0, "secs");
                t0 = time.time()
                fitsfile = None
                images = fitsdata_to_images(fitsdata)
                dprint(3, "converting to images took", time.time() - t0, "secs");
                t0 = time.time()
                fitsdata = None

            data = images[num_image]

            title = self.dp.filename
            if nplanes > 1:
                title += ", plane #%d" % num_image
            Purr.progressMessage("rendering %s" % title, sub=True)

            # min/max data values
            dprint(3, "rendering plane", num_image);
            t0 = time.time()
            datamask = ~numpy.isfinite(data)
            dprint(3, "making mask took", time.time() - t0, "secs");
            t0 = time.time()
            datamin, datamax = scipy.ndimage.measurements.extrema(data, datamask, False)[:2]
            dprint(3, "computing min/max took", time.time() - t0, "secs");
            t0 = time.time()
            rec.datamin, rec.datamax = datamin, datamax
            # mean and sigma
            rec.datamean = scipy.ndimage.measurements.mean(data, datamask, False)
            dprint(3, "computing mean took", time.time() - t0, "secs");
            t0 = time.time()
            rec.datastd = scipy.ndimage.measurements.standard_deviation(data, datamask, False)
            dprint(3, "computing std took", time.time() - t0, "secs");
            t0 = time.time()
            # thumbnail files will be "" if images are small enough to be inlined.
            # these will be None if no histogram clipping is applied
            rec.clipmin, rec.clipmax = None, None
            dprintf(3, "%s plane %d: datamin %g, datamax %g\n", self.dp.fullpath, num_image, rec.datamin, rec.datamax)
            # compute histogram of data only if this is enabled,
            # and either pychart is available (so we can produce plots), or histogram clipping is in effect
            if datamin != datamax and nbins and (pychart or hclip):
                dprintf(3, "%s plane %d: computing histogram\n", self.dp.fullpath, num_image)
                counts = scipy.ndimage.measurements.histogram(data, datamin, datamax, nbins, labels=datamask,
                                                              index=False);  # needed for 1.3+ to avoid warnings
                edges = datamin + (datamax - datamin) * (numpy.arange(nbins, dtype=float) + .5) / nbins
                dprint(3, "computing histogram took", time.time() - t0, "secs");
                t0 = time.time()
                # render histogram
                if pychart:
                    try:
                        self._make_histogram(hf_path, "Histogram of %s" % title, edges, counts)
                        dprint(3, "rendering histogram took", time.time() - t0, "secs");
                        t0 = time.time()
                    except:
                        print(("Error rendering histogram %s" % hf_path))
                        traceback.print_exc()
                        rec.histogram_full = None
                    # if histogram was rendered, make a thumbnail
                    if rec.histogram_full:
                        self.makeThumb(hf_path, hf_thumb, tsize_hist)
                    else:
                        rec.histogram_full_thumb = None
                # now, compute clipped data if needed
                if hclip:
                    # find max point in histogram
                    ic = counts.argmax()
                    # compute number of points that need to be included, given the clip factor
                    target_count = int(data.size * hclip)
                    ih0 = ih1 = ic
                    totcount = counts[ic]
                    # find how many bins to include around ic, stopping when we hit the edge
                    while totcount < target_count:
                        if ih0 > 0:
                            ih0 -= 1
                            totcount += counts[ih0]
                        if ih1 < nbins - 1:
                            ih1 += 1
                            totcount += counts[ih1]
                        # just in case
                        if ih0 <= 0 and ih1 >= nbins - 1:
                            break
                    # and these are the clipping limits
                    datamin = float(edges[ih0])
                    if ih1 >= nbins - 1:
                        ih1 = nbins - 1;  # and datamax is already the clipping limit
                    else:
                        ih1 += 1
                        datamax = float(edges[ih1])
                    rec.clipmin, rec.clipmax = datamin, datamax
                    dprintf(3, "%s plane %d: clipping to %g,%g\n", self.dp.fullpath, num_image, rec.clipmin,
                            rec.clipmax)
                    # render zoomed histogram
                    if pychart:
                        if rec.clipmax != rec.clipmin:
                            zcounts = scipy.ndimage.measurements.histogram(data, rec.clipmin, rec.clipmax, nbins,
                                                                           labels=datamask,
                                                                           index=False);  # needed for 1.3+ to avoid warnings
                            zedges = rec.clipmin + (rec.clipmax - rec.clipmin) * (
                                        numpy.arange(nbins, dtype=float) + .5) / nbins
                            try:
                                self._make_histogram(hz_path, "Histogram zoom of %s" % title, zedges, zcounts)
                                dprint(3, "rendering zoomed histogram took", time.time() - t0, "secs");
                                t0 = time.time()
                            except:
                                print(("Error rendering histogram %s" % hz_path))
                                traceback.print_exc()
                                rec.histogram_zoom = None
                        else:  # no meaningful zoomed area to render
                            rec.histogram_zoom = None
                        # if histogram was rendered, make a thumbnail
                        if rec.histogram_zoom:
                            histogram_zoom_thumb = self.makeThumb(hz_path, hz_thumb, tsize_hist)
                        else:
                            rec.histogram_zoom_thumb = None
                    # clip data
                    data = numpy.clip(data, datamin, datamax)
                # end of clipping
            # else no histogram for whatever reason
            else:
                rec.histogram_full = rec.histogram_zoom = rec.histogram_full_thumb = rec.histogram_zoom_thumb = None
            # ok, data has been clipped if need be. Rescale it to 8-bit integers
            t0 = time.time()
            datarng = datamax - datamin
            if datarng:
                data = (data - datamin) * (255 / datarng)
                data = data.round().astype('uint8')
                data[datamask] = 255
            else:
                data = numpy.zeros(data.shape, dtype='uint8')
            dprintf(3, "%s plane %d: rescaled to %d:%d in %f seconds\n", self.dp.fullpath, num_image, data.min(),
                    data.max(), time.time() - t0);
            t0 = time.time()
            # generate PNG image
            img = None
            try:
                img = PIL.Image.frombuffer('L', data.shape[-1::-1], numpy.getbuffer(data), "raw", 'L', 0, -1)
                dprint(3, "image frombuffer took", time.time() - t0, "secs");
                t0 = time.time()
                # img = PIL.Image.new('L',data.shape)
                # dprint(3,"new image took",time.time()-t0,"secs"); t0 = time.time()
                # imgdata = data.reshape((data.size,))
                # dprint(3,"data.reshape took",time.time()-t0,"secs"); t0 = time.time()
                # img.putdata(imgdata)
                # dprint(3,"putdata took",time.time()-t0,"secs"); t0 = time.time()
                # img = img.transpose(PIL.Image.FLIP_TOP_BOTTOM)
                # dprint(3,"transpose took",time.time()-t0,"secs"); t0 = time.time()
                img.save(img_path, 'PNG')
                dprint(3, "saving took", time.time() - t0, "secs");
                t0 = time.time()
            except:
                print(("Error rendering image %s" % path))
                traceback.print_exc()
                rec.fullimage = img = None
            # if image was rendered, make a thumbnail
            if rec.fullimage:
                thumb = self.makeThumb(img_path, img_thumb, tsize_img, img=img)
                dprint(3, "rendering thumbnail took", time.time() - t0, "secs");
                t0 = time.time()
                # None means thumbnail failed
                if thumb is None:
                    rec.thumbnail = None
                # else perhaps image is its own thumbnail
                elif thumb is img_path:
                    rec.thumbnail = rec.fullimage
            else:
                rec.thumbnail = None
            # write stats
            try:
                pickle.dump(rec, open(recpath, 'rw'))
            except:
                print(("Error writing stats file  %s" % recpath))
                traceback.print_exc()

    def makeThumb(self, imagepath, thumbpath, tsize, img=None):
        """makes a thumbnail for the given image.
        imagepath refers to an image file
        img can be an open PIL.Image -- if None, then imagepath is opened
        tsize is a width,height tuple giving the max thumbnail size
        extension is the extension of the thumbnail file
        """
        try:
            # open image if needed
            if not img:
                img = PIL.Image.open(imagepath)
            # do we need a thumbnail at all, or can the image be inlined?
            width, height = img.size
            factor = max(width / float(tsize[0]), height / float(tsize[1]))
            if factor <= 1:
                return imagepath
            # generate the thumbnail
            img = img.resize((int(width / factor), int(height / factor)), PIL.Image.ANTIALIAS)
            img.save(thumbpath, "PNG")
            return thumbpath
        except:
            print(("Error rendering thumbnail %s" % thumbpath))
            traceback.print_exc()
            return None

    def _renderSingleImage(self, image, thumb, relpath):
        if image is None:
            return ""
        # else no thumbnail: make "image" link
        elif thumb is None:
            fname = relpath + image
            return """<A HREF="%s">image</A>""" % quote_url(fname)
        # else return thumbnail linking to full image
        else:
            tname = relpath + thumb
            fname = relpath + image
            return """<A HREF="%s"><IMG SRC="%s" ALT="%s"></A>""" % (quote_url(fname), quote_url(tname),
                                                                     quote_url(os.path.basename(image)))

    def _renderImageRec(self, rec, relpath, include_size=False):
        # get HTML code for image and histograms
        html_image = self._renderSingleImage(rec.fullimage, rec.thumbnail, relpath)
        if rec.histogram_full:
            html_hist_full = self._renderSingleImage(rec.histogram_full, rec.histogram_full_thumb, relpath)
        else:
            html_hist_full = ""
        if rec.histogram_zoom:
            html_hist_zoom = self._renderSingleImage(rec.histogram_zoom, rec.histogram_zoom_thumb, relpath)
        else:
            html_hist_zoom = ""
        # arrange images side-by-side
        html_img = "<TABLE><TR><TD ROWSPAN=2>%s</TD><TD>%s</TD></TR><TR><TD>%s</TD></TR></TABLE>" % \
                   (html_image, html_hist_full, html_hist_zoom)
        # form up comments
        html_cmt = """<TABLE><TR><TD>data range:</TD><TD>%g,%g</TD></TR>""" % (rec.datamin, rec.datamax)
        if include_size:
            html_cmt += """
         <TR><TD>size:</TD><TD>%s</TD></TR>\n""" % self.cubesize
        html_cmt += """
         <TR><TD>mean:</TD><TD>%g</TD></TR>
         <TR><TD>sigma:</TD><TD>%g</TD></TR>""" % (rec.datamean, rec.datastd)
        if rec.clipmin is not None:
            html_cmt += """
         <TR><TD>clipping:</TD><TD>%g%%</TD></TR>
         <TR><TD>clip range:</TD><TD>%g,%g</TD></TR>""" % (self.hclip * 100, rec.clipmin, rec.clipmax)
        html_cmt += """\n
       </TABLE>"""
        return html_img, html_cmt

    def renderLink(self, relpath=""):
        """renderLink() is called to render a link to the DP
        """
        # return from cache if available
        cachekey, html = self.checkCache('Link', relpath)
        if html is not None:
            return html
        # else regenerate
        html = CachingRenderer.renderLink(self, relpath)
        if self.headerfile is not None:
            html += """ (<A HREF="%s">header</A>)""" % quote_url(relpath + self.headerfile)
        # save to cache
        return self.writeCache(cachekey, html)

    def renderInTable(self, relpath=""):
        """renderInTable() is called to render FITS images in a table"""
        # return from cache if available
        cachekey, html = self.checkCache('InTable', relpath)
        if html is not None:
            return html
        # else regenerate
        # single image: render as standard cells
        if len(self.imgrec) == 1:
            rec = self.imgrec[0]
            # add header
            html = "    <TR><TD COLSPAN=2>"
            html += self.renderLinkComment(relpath) or ""
            html += "</TD></TR>\n"
            html_img, comment = self._renderImageRec(rec, relpath, include_size=True)
            html += "\n".join([
                "    <TR>",
                "      <TD>%s</TD>" % html_img,
                "      <TD>%s</TD>" % comment,
                "    </TR>\n"])
        # multiple images: render a single header row, followed by one row per image
        else:
            # add header
            html = "    <TR><TD COLSPAN=2>"
            html += self.renderLinkComment(relpath)
            # append information on image and end the table row
            html += "\n      <DIV ALIGN=right><P>%s FITS cube, %d planes are given below.</P></DIV></TD></TR>\n" % (
            self.cubesize, len(self.imgrec))
            # now loop over images and generate a table row for each
            for irec, rec in enumerate(self.imgrec):
                html_img, comment = self._renderImageRec(rec, relpath)
                comment = "<P>Image plane #%d.</P>%s" % (irec, comment)
                html += "\n".join([
                    "    <TR>",
                    "      <TD>%s</TD>" % html_img,
                    "      <TD>%s</TD>" % comment,
                    "    </TR>\n"])
        return self.writeCache(cachekey, html)

    def renderThumbnail(self, relpath=""):
        """renderThumbnail() is called to render a thumbnail of the DP.
        We only render the first image (in case of multiple images)
        """
        # return from cache if available
        cachekey, html = self.checkCache('Thumbnail', relpath)
        if html is not None:
            return html
        # else regenerate
        rec = self.imgrec[0]
        html = self._renderSingleImage(rec.fullimage, rec.thumbnail, relpath)
        # save to cache
        return self.writeCache(cachekey, html)


# register ourselves with Purr
import Purr.Render

Purr.Render.addRenderer(FITSRenderer, __name__, __file__)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import basecanvas
from .scaling import *

comment_p = 0

# Converts BMP Unicode character to Postscript character name.  In
# theory, we should be able to just use ISOLatin1Encoding, but this
# doesn't seem to work.
_unicode_latin1_conversion_table = {
    0x0027: 'quotesingle',
    0x0060: 'grave',
    0x00A0: 'space',
    0x00A1: 'exclamdown',
    0x00A2: 'cent',
    0x00A3: 'sterling',
    0x00A4: 'currency',
    0x00A5: 'yen',
    0x00A6: 'brokenbar',
    0x00A7: 'section',
    0x00A8: 'dieresis',
    0x00A9: 'copyright',
    0x00AA: 'ordfeminine',
    0x00AB: 'guillemotleft',
    0x00AC: 'logicalnot',
    0x00AD: 'hyphen',
    0x00AE: 'registered',
    0x00AF: 'macron',
    0x00B0: 'degree',
    0x00B1: 'plusminus',
    0x00B2: 'twosuperior',
    0x00B3: 'threesuperior',
    0x00B4: 'acute',
    0x00B5: 'mu',
    0x00B6: 'paragraph',
    0x00B7: 'periodcentered',
    0x00B8: 'cedilla',
    0x00B9: 'onesuperior',
    0x00BA: 'ordmasculine',
    0x00BB: 'guillemotright',
    0x00BC: 'onequarter',
    0x00BD: 'onehalf',
    0x00BE: 'threequarters',
    0x00BF: 'questiondown',
    0x00C0: 'Agrave',
    0x00C1: 'Aacute',
    0x00C2: 'Acircumflex',
    0x00C3: 'Atilde',
    0x00C4: 'Adieresis',
    0x00C5: 'Aring',
    0x00C6: 'AE',
    0x00C7: 'Ccedilla',
    0x00C8: 'Egrave',
    0x00C9: 'Eacute',
    0x00CA: 'Ecircumflex',
    0x00CB: 'Edieresis',
    0x00CC: 'Igrave',
    0x00CD: 'Iacute',
    0x00CE: 'Icircumflex',
    0x00CF: 'Idieresis',
    0x00D0: 'Eth',
    0x00D1: 'Ntilde',
    0x00D2: 'Ograve',
    0x00D3: 'Oacute',
    0x00D4: 'Ocircumflex',
    0x00D5: 'Otilde',
    0x00D6: 'Odieresis',
    0x00D7: 'multiply',
    0x00D8: 'Oslash',
    0x00D9: 'Ugrave',
    0x00DA: 'Uacute',
    0x00DB: 'Ucircumflex',
    0x00DC: 'Udieresis',
    0x00DD: 'Yacute',
    0x00DE: 'Thorn',
    0x00DF: 'germandbls',
    0x00E0: 'agrave',
    0x00E1: 'aacute',
    0x00E2: 'acircumflex',
    0x00E3: 'atilde',
    0x00E4: 'adieresis',
    0x00E5: 'aring',
    0x00E6: 'ae',
    0x00E7: 'ccedilla',
    0x00E8: 'egrave',
    0x00E9: 'eacute',
    0x00EA: 'ecircumflex',
    0x00EB: 'edieresis',
    0x00EC: 'igrave',
    0x00ED: 'iacute',
    0x00EE: 'icircumflex',
    0x00EF: 'idieresis',
    0x00F0: 'eth',
    0x00F1: 'ntilde',
    0x00F2: 'ograve',
    0x00F3: 'oacute',
    0x00F4: 'ocircumflex',
    0x00F5: 'otilde',
    0x00F6: 'odieresis',
    0x00F7: 'divide',
    0x00F8: 'oslash',
    0x00F9: 'ugrave',
    0x00FA: 'uacute',
    0x00FB: 'ucircumflex',
    0x00FC: 'udieresis',
    0x00FD: 'yacute',
    0x00FE: 'thorn',
    0x00FF: 'ydieresis',
}


class T(basecanvas.T):
    def __init__(self, fname):
        basecanvas.T.__init__(self)
        self.__out_fname = fname
        self.__reset_context()
        self.__output_lines = []
        self.__nr_gsave = 0
        self.__font_ids = {}
        self.__nr_fonts = 0

    def __reset_context(self):
        self.__font_name = None
        self.__font_size = -1
        self.__line_style = None
        self.__color = None
        self.__mtx_pushed = 0
        self.__txtmtx_pushed = 0

    def __intern_font(self, name):
        if name in self.__font_ids:
            return self.__font_ids[name]
        id = "F%d" % self.__nr_fonts
        self.__nr_fonts += 1
        self.__font_ids[name] = id
        return id

    def newpath(self):
        self.__write("N\n")

    def stroke(self):
        self.__write("ST\n")

    def closepath(self):
        self.__write("CP\n")

    def moveto(self, x, y):
        self.__write('%g %g M\n' % (x, y))

    def set_fill_color(self, color):
        if self.__color == color:
            pass
        else:
            if color.r == color.g and color.r == color.b:
                self.__write("%g SG\n" % color.r)
            else:
                self.__write("%g %g %g SC\n" % (color.r, color.g, color.b))
            self.__color = color

    def set_stroke_color(self, color):
        self.set_fill_color(color)

    def set_line_style(self, style):
        self.set_stroke_color(style.color)
        if (self.__line_style == style):
            pass
        else:
            self.__write("%g %d %d " % (nscale(style.width),
                                        style.cap_style, style.join_style))
            if style.dash != None:
                self.__write("[%s] 0 SLD " %
                             " ".join(map(str, nscale_seq(style.dash))))
            else:
                self.__write("SL ")
        self.__line_style = style

    def gsave(self):
        self.__nr_gsave += 1
        self.__write("GS\n")

    def grestore(self):
        self.__write("GR\n")
        self.__nr_gsave -= 1
        self.__reset_context()

    def clip_sub(self):
        self.__write("clip\n")

    def path_arc(self, x, y, radius, ratio, start_angle, end_angle):
        self.push_transformation((x, y), (1, ratio), None)
        self.__write("0 0 %g %g %g arc\n" % (radius, start_angle, end_angle))
        self.pop_transformation()

    def curveto(self, a, b, c, d, e, f):
        self.__write("%g %g %g %g %g %g curveto\n" % (a, b, c, d, e, f))

    def push_transformation(self, baseloc, scale, angle, in_text=0):
        self.__mtx_pushed += 1
        self.__write("GB\n")
        if baseloc != None:
            self.__write("%g %g T\n" % (baseloc[0], baseloc[1]))
        if angle != None and angle != 0:
            self.__write("%g R\n" % (angle))
        if scale != None:
            self.__write("%g %g scale\n" % (scale[0], scale[1]))

    def pop_transformation(self, in_text=0):
        if self.__mtx_pushed == 0:
            raise ValueError("mtx not pushed")
        self.__mtx_pushed -= 1
        self.__write("GE\n")

    def text_begin(self):
        self.__txtmtx_pushed += 1
        self.__write("TB\n")

    def text_end(self):
        self.__txtmtx_pushed -= 1
        self.__write("TE\n")

    def text_moveto(self, x, y, angle):
        self.__write("%g %g T " % (x, y))
        if angle != None and angle != 0:
            self.__write("%g R " % angle)
        self.moveto(0, 0)

    def text_show(self, font_name, size, color, str):
        self.set_fill_color(color)
        if (self.__font_name == font_name and self.__font_size == size):
            pass
        else:
            font_id = self.__intern_font(font_name)
            self.__write("%g %s\n" % (size, font_id))
            self.__font_name = font_name
            self.__font_size = size

        self.__write("(")
        # Convert unicode to Postscript Latin code.    
        for ch in str:
            # Note: Escaping of ()\  is done by the basecanvas.show, so
            # we need not worry about them here. Perhaps we should move
            # that code here??
            n = ord(ch)
            if n > 128 or ch in ("'", "`"):
                ent = _unicode_latin1_conversion_table.get(n, None)
                if not ent:
                    self.__write(ch)
                else:
                    self.__write(") show /%s glyphshow (" % ent)
            else:
                self.__write(ch)
        self.__write(") show\n")

    def _path_polygon(self, points):
        if (len(points) == 4
                and points[0][0] == points[1][0]
                and points[2][0] == points[3][0]
                and points[0][1] == points[3][1]
                and points[1][1] == points[2][1]):
            # a rectangle.
            (xmin, ymin, xmax, ymax) = basecanvas._compute_bounding_box(points)
            if basecanvas.invisible_p(xmax, ymax):
                return
            self.setbb(xmin, ymin)
            self.setbb(xmax, ymax)
            self.__write("%g %g %g %g RECT\n" % \
                         (xscale(points[0][0]), yscale(points[0][1]),
                          xscale(points[2][0]), yscale(points[2][1])))
        else:
            basecanvas.T._path_polygon(self, points)

    def lineto(self, x, y):
        self.__write("%g %g L\n" % (x, y))

    def fill(self):
        self.__write("fill\n")

    def comment(self, str):
        if comment_p:
            self.verbatim("%" + str)

    def verbatim(self, str):
        self.__write(str)

    def close(self):
        basecanvas.T.close(self)
        if self.__output_lines == []:
            return

        fp, need_close = self.open_output(self.__out_fname)

        if self.__nr_gsave != 0:
            raise Exception("gsave misnest (%d)" % (self.__nr_gsave))
        self.write_preamble(fp)

        fp.writelines(self.__output_lines)
        fp.writelines(["showpage end\n",
                       "%%Trailer\n",
                       "%%EOF\n"])
        if need_close:
            fp.close()

    def __write(self, str):
        self.__output_lines.append(str)

    def writelines(self, l):
        self.__output_lines.extend(l)

    def write_preamble(self, fp):
        bbox = [self.__xmin - 1, self.__ymin - 1, self.__xmax + 1, self.__ymax + 1]
        fp.write("%!PS-Adobe-2.0 EPSF-1.2\n")
        fp.write("%%Title: " + self.title + "\n")
        fp.write("%%Creator: " + self.creator + "\n")
        if self.author:
            fp.write("%%Author: " + self.author + "\n")
        fp.write("%%CreationDate: " + self.creation_date + "\n")
        fp.write("%%DocumentFonts: " + " ".join(list(self.__font_ids.keys())) + "\n")
        fp.write("%%Pages: 1\n")

        bbox = theme.adjust_bounding_box(bbox)

        fp.write("%%%%BoundingBox: %d %d %d %d\n" % \
                 (round(xscale(bbox[0])),
                  round(yscale(bbox[1])),
                  round(xscale(bbox[2])),
                  round(yscale(bbox[3]))))
        fp.write("%%EndComments\n")
        if self.aux_comments != "":
            for line in self.aux_comments.split("\n"):
                fp.write("% " + line + "\n")

        fp.write(preamble_text)
        for name, font_id in list(self.__font_ids.items()):
            fp.write("/%s {/%s findfont SF} def\n" % (font_id, name))
        fp.write("%%EndProlog\n%%Page: 1 1\n")


preamble_text = """
40 dict begin
/RECT {4 dict begin
  /y2 exch def
  /x2 exch def
  /y1 exch def
  /x1 exch def
  newpath x1 y1 moveto x2 y1 lineto x2 y2 lineto x1 y2 lineto closepath
  end
} def

/SF {exch scalefont setfont} def
/TB {matrix currentmatrix} def
/TE {setmatrix} def
/GB {matrix currentmatrix} def
/GE {setmatrix} def
/SG {1 1 1 setrgbcolor setgray} def
/SC {1 setgray setrgbcolor} def
/SL {[] 0 setdash setlinejoin setlinecap setlinewidth} def
/SLD {setdash setlinejoin setlinecap setlinewidth} def
/M {moveto} def
/L {lineto} def
/T {translate} def
/R {rotate} def
/N {newpath} def
/ST {stroke} def
/CP {closepath} def
/GR {grestore} def
/GS {gsave} def
"""

# SL: set line style.
# width [dash] x linecap linejoin SL -> 

# SF: set font.
# name size SF ->



# -*- coding: utf-8 -*-
import os
import os.path
import time

import six
import Kittens.utils
import Kittens.widgets
from PyQt4.Qt import (QWidget, QDialog, QWizard, QWizardPage, QButtonGroup, QVBoxLayout, QRadioButton, QObject, SIGNAL,
                      QHBoxLayout, QLineEdit, QPushButton, QFileDialog, QMessageBox, QHeaderView, QAbstractItemView,
                      QFontMetrics, QFont, QTreeWidget, QSizePolicy, QMenu, QMimeData, QUrl, QComboBox, QMimeData,
                      QTreeWidgetItem, Qt, QApplication, QClipboard, QLabel, QSplitter, QTextEdit,
                      QTextDocument, QSize, QFrame, QStackedWidget, QWidgetAction, QMenu, QTextBrowser, QPoint, QDrag)
if six.PY3:
    QVariant = str
else:
    from PyQt4.Qt import QVariant

import Purr.LogEntry
import Purr.Render
from Purr import Config, pixmaps, dprint


def _makeUniqueFilename(taken_names, name):
    """Helper function. Checks if name is in the set 'taken_names'.
    If so, attepts to form up an untaken name (by adding numbered suffixes).
    Adds name to taken_names.
    """
    if name in taken_names:
        # try to form up new name
        basename, ext = os.path.splitext(name)
        num = 1
        name = "%s-%d%s" % (basename, num, ext)
        while name in taken_names:
            num += 1
            name = "%s-%d%s" % (basename, num, ext)
    # finally, enter name into set
    taken_names.add(name)
    return name


if six.PY3:
    maketrans = str.maketrans
else:
    import string
    maketrans = string.maketrans

_sanitize_chars = "\\:*?\"<>|"
_sanitize_trans = maketrans(_sanitize_chars, '_' * len(_sanitize_chars))


def _sanitizeFilename(filename):
    """Sanitizes filename for use on Windows and other brain-dead systems, by replacing a number of illegal characters
    with underscores."""
    global _sanitize_trans
    out = str(filename).translate(_sanitize_trans)
    # leading dot becomes "_"
    if out and out[0] == '.':
        out = out[1:]
    return out


class DPTreeWidget(Kittens.widgets.ClickableTreeWidget):
    """This class implements a QTreeWidget for data products.
    """

    def __init__(self, *args):
        Kittens.widgets.ClickableTreeWidget.__init__(self, *args)
        # insert columns, and numbers for them
        self.header().setDefaultSectionSize(120)
        columns = ["action", "filename", "type", "rename to", "render", "comment"]
        self.setHeaderLabels(columns)
        self.ColAction, self.ColFilename, self.ColType, self.ColRename, self.ColRender, self.ColComment = list(
            range(len(columns)))
        self.setSortingEnabled(False)
        self.setRootIsDecorated(False)
        self.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.setDragEnabled(True)
        # sort out resizing modes
        self.header().setMovable(False)
        self.header().setResizeMode(self.ColFilename, QHeaderView.Interactive)
        self.header().setResizeMode(self.ColType, QHeaderView.ResizeToContents)
        self.header().setResizeMode(self.ColRename, QHeaderView.Interactive)
        self.header().setResizeMode(self.ColComment, QHeaderView.Stretch)
        self.header().setResizeMode(self.ColAction, QHeaderView.Fixed)
        self.header().setResizeMode(self.ColRender, QHeaderView.Fixed)
        self.header().resizeSection(self.ColAction, 64)
        self.header().resizeSection(self.ColRender, 64)
        self._fontmetrics = QFontMetrics(QFont())
        # setup other properties of the listview
        self.setAcceptDrops(True)
        try:
            self.setAllColumnsShowFocus(True)
        except AttributeError:
            pass;  # qt 4.2+
        # self.setDefaultRenameAction(QTreeWidget.Accept)
        self.setSelectionMode(QTreeWidget.SingleSelection)
        self.header().show()
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        self.connect(self, SIGNAL("mouseButtonClicked"), self._itemClicked)
        self.connect(self, SIGNAL("itemActivated(QTreeWidgetItem*,int)"), self._itemActivated)
        self.connect(self, SIGNAL("itemChanged(QTreeWidgetItem*,int)"), self._itemRenamed)
        self.connect(self, SIGNAL("currentItemChanged(QTreeWidgetItem*,QTreeWidgetItem*)"), self._currentItemChanged)
        self.connect(self, SIGNAL("itemcontextMenuRequested"), self._showItemContextMenu)
        self.connect(self, SIGNAL("droppedFiles"), self, SIGNAL("filesSelected"))
        # create popup menu for existing DPs
        self._archived_dp_menu = menu = QMenu()
        menu.addAction(pixmaps.editcopy.icon(), "Restore file from this entry's archived copy",
                       self._restoreItemFromArchive)
        menu.addAction(pixmaps.editpaste.icon(), "Copy location of archived copy to clipboard",
                       self._copyItemToClipboard)
        # currently selected item
        self._current_item = None
        # item onto which something was dropped
        self._dropped_on = None
        # currently edited item and column
        self._editing = None
        # dictionary of listview items
        # key is (path,archived) tuple, value is item
        self.dpitems = {}

        self._policy_list_default = (("copy", pixmaps.copy.icon()),
                                     ("move", pixmaps.move.icon()),
                                     ("ignore", pixmaps.grey_round_cross.icon()),
                                     ("banish", pixmaps.red_round_cross.icon()))
        self._policy_list_archived = (("keep", pixmaps.checkmark.icon()),
                                      ("remove", pixmaps.grey_round_cross.icon()))

    def _checkDragDropEvent(self, ev):
        """Checks if event contains a file URL, accepts if it does, ignores if it doesn't"""
        mimedata = ev.mimeData()
        if mimedata.hasUrls():
            urls = [str(url.toLocalFile()) for url in mimedata.urls() if url.toLocalFile()]
        else:
            urls = []
        # accept event if drag text is a file URL
        if urls:
            ev.acceptProposedAction()
            return urls
        else:
            ev.ignore()
            return None

    def dragEnterEvent(self, ev):
        """Process drag-enter event. Use function above to accept or ignore it"""
        self._checkDragDropEvent(ev)

    def dragMoveEvent(self, ev):
        """Process drag-move event. Use function above to accept or ignore it"""
        self._checkDragDropEvent(ev)

    def dropEvent(self, ev):
        """Process drop event."""
        # use function above to accept event if it contains a file URL
        files = self._checkDragDropEvent(ev)
        if files:
            pos = ev.pos()
            dropitem = self.itemAt(pos)
            dprint(1, "dropped on", pos.x(), pos.y(), dropitem and str(dropitem.text(1)))
            # if event originated with ourselves, reorder items
            if ev.source() is self:
                self.reorderItems(dropitem, *files)
            # else event is from someone else, accept the dropped files
            else:
                self._dropped_on = dropitem
                self.emit(SIGNAL("droppedFiles"), *files)
                # if event originated with another DPTreeWidget, emit a draggedAwayFiles() signal on its behalf
                if isinstance(ev.source(), DPTreeWidget):
                    ev.source().emit(SIGNAL("draggedAwayFiles"), *files)

    def reorderItems(self, beforeitem, *files):
        # make list of items to be moved
        moving_items = []
        for ff in files:
            item = self.dpitems.get(ff, None)
            # if dropping item on top of itself, ignore the operation
            if item is beforeitem:
                return
            if item:
                dum = QWidget()
                for col in self.ColAction, self.ColRender:
                    widget = self.itemWidget(item, self.ColRender)
                    widget and widget.setParent(dum)
                moving_items.append(self.takeTopLevelItem(self.indexOfTopLevelItem(item)))
        dprint(1, "moving items", [str(item.text(1)) for item in moving_items])
        dprint(1, "to before", beforeitem and str(beforeitem.text(1)))
        # move them to specified location
        if moving_items:
            index = self.indexOfTopLevelItem(beforeitem) if beforeitem else self.topLevelItemCount()
            self.insertTopLevelItems(index, moving_items)
            for item in moving_items:
                for col in self.ColAction, self.ColRender:
                    self._itemComboBox(item, col)
            self.emit(SIGNAL("updated"))

    def mimeTypes(self):
        return ["text/x-url"]

    def mimeData(self, itemlist):
        mimedata = QMimeData()
        urls = []
        for item in itemlist:
            if item._dp:
                urls.append(QUrl.fromLocalFile(item._dp.fullpath or item._dp.sourcepath))
        mimedata.setUrls(urls)
        return mimedata

    def clear(self):
        QTreeWidget.clear(self)
        self.dpitems = {}

    def _itemComboBox(self, item, column):
        """This returns the QComboBox associated with item and column, or creates a new one if
        it hasn't been created yet. The reason we don't create a combobox immediately is because
        the item needs to be inserted into its QTreeWidget first."""
        if item.treeWidget() is not self:
            return None
        combobox = self.itemWidget(item, column)
        if not combobox:
            combobox = QComboBox(self)
            # combobox.setSizeAdjustPolicy(QComboBox.AdjustToContents)
            self.setItemWidget(item, column, combobox)
            width = 0
            for num, (name, icon) in enumerate(item._combobox_option_list[column]):
                if icon:
                    combobox.addItem(icon, name)
                else:
                    combobox.addItem(name)
                width = max(width, self._fontmetrics.width(name) + (icon and 16 or 0))
            combobox.setCurrentIndex(item._combobox_current_index[column])
            item._combobox_changed[column] = Kittens.utils.curry(self._updateItemComboBoxIndex, item, column)
            QObject.connect(combobox, SIGNAL("currentIndexChanged(int)"), item._combobox_changed[column])
            QObject.connect(combobox, SIGNAL("currentIndexChanged(int)"), self._emitUpdatedSignal)
            # resize section if needed
            width += 32
            if width > self.header().sectionSize(column):
                self.header().resizeSection(column, width)
        return combobox

    def _emitUpdatedSignal(self, *dum):
        self.emit(SIGNAL("updated"))

    def _updateItemComboBoxIndex(self, item, column, num):
        """Callback for comboboxes: notifies us that a combobox for the given item and column has changed"""
        item._combobox_current_index[column] = num
        item._combobox_current_value[column] = item._combobox_option_list[column][num][0]

    def setItemPolicy(self, item, policy):
        """Sets the policy of the given item"""
        index = item._combobox_indices[self.ColAction].get(policy, 0)
        self._updateItemComboBoxIndex(item, self.ColAction, index)
        combobox = self.itemWidget(item, self.ColAction)
        if combobox:
            combobox.setCurrentIndex(index)

    def getItemDPList(self):
        """Returns list of item,dp pairs corresponding to content of listview.
        Not-yet-saved items will have dp=None."""
        itemlist = [(item, item._dp) for item in self.iterator()]
        return itemlist

    def _makeDPItem(self, parent, dp, after=None):
        """Creates listview item for data product 'dp', inserts it after item 'after'"""
        if parent:
            item = QTreeWidgetItem(parent, after)
        else:
            item = QTreeWidgetItem()
        item.setTextAlignment(self.ColAction, Qt.AlignRight | Qt.AlignVCenter)
        item.setTextAlignment(self.ColFilename, Qt.AlignRight | Qt.AlignVCenter)
        item.setTextAlignment(self.ColType, Qt.AlignLeft | Qt.AlignVCenter)
        item.setTextAlignment(self.ColRename, Qt.AlignLeft | Qt.AlignVCenter)
        item.setTextAlignment(self.ColRender, Qt.AlignHCenter | Qt.AlignVCenter)
        item.setTextAlignment(self.ColComment, Qt.AlignLeft | Qt.AlignVCenter)
        item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsDragEnabled | Qt.ItemIsDropEnabled | Qt.ItemIsEnabled)
        item._dp = dp
        # init stuff for combobox functions above
        item._combobox_option_list = {}
        item._combobox_current_index = {}
        item._combobox_current_value = {}
        item._combobox_indices = {}
        item._combobox_changed = {}
        # setup available policies for new or archived items, set initial policy
        if dp.archived:
            item._menu = self._archived_dp_menu
            item._combobox_option_list[self.ColAction] = policies = self._policy_list_archived
            policy = "keep"
        else:
            item._combobox_option_list[self.ColAction] = policies = self._policy_list_default
            policy = dp.policy
        # create reverse mapping from policy names to indices
        item._combobox_indices[self.ColAction] = dict([(name, num) for num, (name, icon) in enumerate(policies)])
        # init item policy
        self.setItemPolicy(item, policy)
        # set other columns
        basename = os.path.basename(dp.sourcepath)
        name, ext = os.path.splitext(basename)
        item.setText(self.ColFilename, name)
        item.setText(self.ColType, ext)
        item.setToolTip(self.ColFilename, basename)
        item.setToolTip(self.ColType, basename)
        item.setData(self.ColComment, Qt.EditRole, str(dp.comment or ""))
        # make sure new filenames are unique
        filename = _sanitizeFilename(dp.filename)
        if not dp.archived:
            # tack on .tgz prefix onto dirs
            if os.path.isdir(dp.sourcepath) and not filename.endswith(".tgz"):
                filename += ".tgz"
            # form up set of taken names
            taken_names = set()
            for i0, dp0 in self.getItemDPList():
                if dp0.policy not in ["remove", "ignore", "banish"]:
                    taken_names.add(str(i0.text(self.ColRename)))
            # ensure uniqueness of filename
            filename = _makeUniqueFilename(taken_names, filename)
        item.setData(self.ColRename, Qt.EditRole, str(filename))
        # get list of available renderers
        item._renderers = Purr.Render.getRenderers(dp.fullpath or dp.sourcepath)
        item._render = 0
        item._combobox_option_list[self.ColRender] = [(name, None) for name in item._renderers]
        # create reverse mapping from renderer names to indices
        item._combobox_indices[self.ColRender] = dict([(name, num) for num, name in enumerate(item._renderers)])
        # for archived items, try to find renderer in list
        if dp.archived:
            try:
                item._render = item._renderers.index(dp.render)
            except:
                pass
        self._updateItemComboBoxIndex(item, self.ColRender, item._render)
        # add to map of items
        self.dpitems[dp.fullpath or dp.sourcepath] = item
        return item

    def focusOutEvent(self, ev):
        """Redefine focusOut events to stop editing"""
        Kittens.widgets.ClickableTreeWidget.focusOutEvent(self, ev)
        # if focus is going to a child of ours, do nothing
        wid = QApplication.focusWidget()
        while wid:
            if wid is self:
                return
            wid = wid.parent()
        # else we're truly losing focus -- stop the editor
        self._startOrStopEditing()

    def keyPressEvent(self, ev):
        """Stop editing if enter is pressed"""
        if ev.key() in (Qt.Key_Enter, Qt.Key_Return):
            self._startOrStopEditing()
        elif ev.key() == Qt.Key_Escape:
            self._cancelEditing()
        else:
            Kittens.widgets.ClickableTreeWidget.keyPressEvent(self, ev)

    def _cancelEditing(self):
        dprint(2, "cancelling editor")
        if not self._editing:
            return
        item0, column0 = self._editing
        dprint(2, "cancelling editor for", item0.text(1), column0)
        self.closePersistentEditor(*self._editing)
        self._editing = None
        item0.setText(column0, self._editing_oldtext)

    def _startOrStopEditing(self, item=None, column=None):
        if self._editing:
            if self._editing == (item, column):
                return
            else:
                item0, column0 = self._editing
                dprint(2, "closing editor for", item0.text(1), column0)
                self.closePersistentEditor(*self._editing)
                self._editing = None
                if column0 == self.ColRename:
                    item0.setText(self.ColRename, _sanitizeFilename(str(item0.text(self.ColRename))))
        if item and column in [self.ColRename, self.ColComment]:
            self._editing = item, column
            dprint(2, "opening editor for", item.text(1), column)
            self._editing_oldtext = item.text(column)
            self.setCurrentItem(item, column)
            self.openPersistentEditor(item, column)
            self.itemWidget(item, column).setFocus(Qt.OtherFocusReason)

    def _itemActivated(self, item, col):
        dprint(2, "_itemActivated", item.text(1), col)
        self._startOrStopEditing(item, col)

    def _itemClicked(self, button, item, point, column):
        self._startOrStopEditing();  # stop editing
        if not item or button != Qt.LeftButton:
            return
        self._startOrStopEditing(item, column);  # start editing if editable column

    def _itemRenamed(self, item, col):
        if col == self.ColRename:
            item.setText(col, _sanitizeFilename(str(item.text(col))))
        self._editing = None
        self.emit(SIGNAL("updated"))

    def _currentItemChanged(self, item, previous):
        self._startOrStopEditing()

    def _showItemContextMenu(self, item, point, col):
        """Callback for contextMenuRequested() signal. Pops up item menu, if defined"""
        self._startOrStopEditing()
        menu = getattr(item, '_menu', None)
        if menu:
            # self._current_item tells callbacks what item the menu was referring to
            self._current_item = item
            self.clearSelection()
            self.setItemSelected(item, True)
            menu.exec_(point)
        else:
            self._current_item = None

    def _copyItemToClipboard(self):
        """Callback for item menu."""
        dp = self._current_item and getattr(self._current_item, '_dp', None)
        if dp and dp.archived:
            path = dp.fullpath.replace(" ", "\\ ")
            QApplication.clipboard().setText(path, QClipboard.Clipboard)
            QApplication.clipboard().setText(path, QClipboard.Selection)

    def _restoreItemFromArchive(self):
        """Callback for item menu."""
        dp = self._current_item and getattr(self._current_item, '_dp', None)
        if dp and dp.archived:
            dp.restore_from_archive(parent=self)

    def fillDataProducts(self, dps):
        """Fills listview with existing data products"""
        item = None
        for dp in dps:
            if not dp.ignored:
                item = self._makeDPItem(self, dp, item)
                # ensure combobox widgets are made
                self._itemComboBox(item, self.ColAction)
                self._itemComboBox(item, self.ColRender)

    def addDataProducts(self, dps):
        """Adds new data products to listview. dps is a list of DP objects.
        Returns True if new (non-quiet) DPs are added, or if existing non-quiet dps are updated.
        (this usually tells the main window to wake up)
        """
        busy = Purr.BusyIndicator()
        wakeup = False
        # build up list of items to be inserted
        itemlist = []
        for dp in dps:
            item = self.dpitems.get(dp.sourcepath)
            # If item already exists, it needs to be moved to its new position
            # If takeTopLevelItem() returns None, then item was already removed (this shouldn't happen,
            # but let's be defensive), and we make a new one anyway.
            if item and self.takeTopLevelItem(self.indexOfTopLevelItem(item)):
                itemlist.append(item)
            else:
                itemlist.append(self._makeDPItem(None, dp))
            wakeup = wakeup or not (dp.ignored or dp.quiet)
        # if these DPs were added as a result of a drag-and-drop, we need to insert them in FRONT of the dropped-on item
        if self._dropped_on:
            index = self.indexOfTopLevelItem(self._dropped_on)
        # else insert at end (after=None)
        else:
            index = self.topLevelItemCount()
        if itemlist:
            self.insertTopLevelItems(index, itemlist)
            self.emit(SIGNAL("updated"))
        for item in itemlist:
            # ensure combobox widgets are made
            self._itemComboBox(item, self.ColAction)
            self._itemComboBox(item, self.ColRender)
        return wakeup

    def dropDataProducts(self, *pathnames):
        """Drops (that is, deletes) new (i.e. non-archived) DP items matching the given pathnames."""
        trash = QTreeWidget(None)
        updated = False
        for path in pathnames:
            item = self.dpitems.get(path)
            if item and not item._dp.archived:
                self.takeTopLevelItem(self.indexOfTopLevelItem(item))
                trash.addTopLevelItem(item)
                del self.dpitems[path]
                updated = True
        if updated:
            self.emit(SIGNAL("updated"))

    def resolveFilenameConflicts(self):
        """Goes through list of DPs to make sure that their destination names
        do not clash. Adjust names as needed. Returns True if some conflicts were resolved.
        """
        taken_names = set()
        resolved = False
        # iterate through items
        for item, dp in self.getItemDPList():
            # only apply this to saved DPs
            if dp.policy not in ["remove", "ignore", "banish"]:
                name0 = str(item.text(self.ColRename))
                name = _makeUniqueFilename(taken_names, name0)
                if name != name0:
                    item.setText(self.ColRename, name)
                    resolved = True
                    self.emit(SIGNAL("updated"))
        return resolved

    def buildDPList(self):
        """Builds list of data products."""
        updated = False
        dps = []
        itemlist = self.getItemDPList()
        # first remove all items marked for removal, in case their names clash with new or renamed items
        for item, dp in itemlist:
            item._policy = item._combobox_current_value[self.ColAction]
            if dp and item._policy == "remove":
                dp.remove_file()
                dp.remove_subproducts()
        # now, make list
        for item, dp in itemlist:
            if item._policy == "remove":
                continue
            # update renderer and comment
            render = item._combobox_current_value[self.ColRender]
            comment = str(item.text(self.ColComment))
            if render != dp.render:
                dp.render = render
                updated = True
            if comment != dp.comment:
                dp.comment = comment
                updated = True
            # archived DPs may need to be renamed, for new ones simply set the name
            if dp.archived:
                updated = updated or dp.rename(str(item.text(self.ColRename)))
            else:
                dp.set_policy(item._policy)
                dp.filename = str(item.text(self.ColRename))
            dps.append(dp)
        return dps, updated


class LogEntryEditor(QWidget):
    """This class provides a widget for editing log entries.
    """

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        # create splitter
        lo = QVBoxLayout(self)
        lo.setMargin(0)
        self.wsplitter = QSplitter(self)
        self.wsplitter.setOrientation(Qt.Vertical)
        self.wsplitter.setChildrenCollapsible(False)
        # self.wsplitter.setFrameStyle(QFrame.Box|QFrame.Raised)
        # self.wsplitter.setMargin(0)
        self.wsplitter.setLineWidth(0)
        lo.addWidget(self.wsplitter)
        # create pane for comment editor
        editorpane = QWidget(self.wsplitter)
        lo_top = QVBoxLayout(editorpane)
        # lo_top.setResizeMode(QLayout.Minimum)
        lo_top.setMargin(0)
        # create comment editor
        # create title and timestamp label, hide timestamp until it is set (below)
        lo_topline = QHBoxLayout();
        lo_top.addLayout(lo_topline)
        lo_topline.setMargin(0)
        self.wtoplabel = QLabel("Entry title:", editorpane)
        self.wtimestamp = QLabel("", editorpane)
        lo_topline.addWidget(self.wtoplabel)
        lo_topline.addStretch(1)
        lo_topline.addWidget(self.wtimestamp)
        # add title editor
        self.wtitle = QLineEdit(editorpane)
        lo_top.addWidget(self.wtitle)
        self.connect(self.wtitle, SIGNAL("textChanged(const QString&)"), self._titleChanged)
        # add comment editor
        # lo_top.addSpacing(5)
        lo_top.addWidget(QLabel("Comments:", editorpane))
        self.wcomment = QTextEdit(editorpane)
        self.wcomment.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.comment_doc = QTextDocument(self)
        self.wcomment.setDocument(self.comment_doc)
        self.wcomment.setAcceptRichText(False)
        self.wcomment.setLineWrapMode(QTextEdit.WidgetWidth)
        self.connect(self.comment_doc, SIGNAL("contentsChanged()"), self._commentChanged)
        lo_top.addWidget(self.wcomment)
        # generate frame for the "Data products" listview
        # lo_top.addSpacing(5)
        # create pane for comment editor
        dppane = QWidget(self.wsplitter)
        lo_top = QVBoxLayout(dppane)
        # lo_top.setResizeMode(QLayout.Minimum)
        lo_top.setMargin(0)
        dpline = QWidget(dppane)
        lo_top.addWidget(dpline)
        dpline.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        lo_dpline = QHBoxLayout(dpline)
        lo_dpline.setMargin(0)
        label = QLabel("<nobr>Data products (<u><font color=blue>help</font></u>):</nobr>", dpline)
        label.setToolTip(self.data_product_help)
        lo_dpline.addWidget(label)
        lo_dpline.addStretch(1)
        wnewdp = QPushButton(pixmaps.folder_open.icon(), "Add file...", dpline)
        self.connect(wnewdp, SIGNAL("clicked()"), self._showAddFileDialog)
        wnewdp.setAutoDefault(False)
        wnewdp_dir = QPushButton(pixmaps.folder_open.icon(), "Add dir...", dpline)
        self.connect(wnewdp_dir, SIGNAL("clicked()"), self._showAddDirDialog)
        wnewdp_dir.setAutoDefault(False)
        self._add_dp_dialog = None
        lo_dpline.addWidget(wnewdp)
        lo_dpline.addWidget(wnewdp_dir)
        # create DP listview
        ndplv = self.wdplv = DPTreeWidget(dppane)
        ndplv.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        lo_top.addWidget(self.wdplv)
        # connect its signals
        QObject.connect(self.wdplv, SIGNAL("updated"), self.setUpdated)
        QObject.connect(self.wdplv, SIGNAL("droppedFiles"), self, SIGNAL("filesSelected"))
        QObject.connect(self.wdplv, SIGNAL("draggedAwayFiles"), self, SIGNAL("draggedAwayFiles"))
        # other internal init
        self.reset()

    data_product_help = \
        """<P><I>Data products</I> are files that will be archived along with this log entry. If "pounce"
        is enabled, PURR will watch your directories for new or updated files, and insert them in this list
        automatically. You can also click on "Add..." to add files by hand.</P>
  
        <P>Click on the <I>action</I> column to select what to do with a file. Click on the <I>rename</I> column
        to archive the file under a different name. Click on the <I>comment</I> column to enter a comment for the file. Click on the <I>render</I> column to select a rendering for the file. The basic rendering is "link", which makes a simple HTML link to the archived file. Certain file types (e.g. images) may support
        more elaborate renderings.</P>
  
        <P>The meaning of the different actions is:</P>
        <DL>
        <DT><B>copy</B></DT><DD>data product will be copied to the log.</DD>
        <DT><B>move</B></DT><DD>data product will be moved to the log.</DD>
        <DT><B>ignore</B></DT><DD>data product will not appear in the log, however PURR will keep
        watching it, and will pounce on it if it changes again.</DD>
        <DT><B>banish</B></DT><DD>data product will not appear in the log, and PURR will stop
        watching it (you can un-banish a data product later by adding it via the
        "Add..." button.)</DD>
        <DT><B>keep</B></DT><DD>(for existing products only) retain data product with this log entry.</DD>
        <DT><B>remove</B></DT><DD>(for existing products only) remove data product from this log entry.</DD>
        </DL>
        """

    def hideEvent(self, event):
        QWidget.hideEvent(self, event)
        if self._add_dp_dialog:
            self._add_dp_dialog.hide()

    def resetDPs(self):
        self.wdplv.clear()

    def reset(self):
        self.wdplv.clear()
        self.entry = self._timestamp = None
        self.wtimestamp.hide()
        self.setEntryTitle("Untitled entry")
        self.setEntryComment("No comment")
        self._default_dir = "."
        # _title_changed is True if title was changed or edited manually
        # _comment_changed is True if comment was changed or edited manually
        # _commend_edited is True if comment was eddited since the last time something was
        #   added with addComment()
        self._title_changed = self._comment_changed = self._comment_edited = False
        # _last_auto_comment is the last comment to have been added via addComment()
        self._last_auto_comment = None
        self.updated = False

    def setUpdated(self, updated=True):
        self.updated = updated
        if updated:
            self.emit(SIGNAL("updated"))

    def setDefaultDirs(self, *dirnames):
        self._default_dirs = dirnames

    class AddDataProductDialog(QFileDialog):
        """This is a file selection dialog with an extra quick-jump combobox for
        multiple directories"""

        def __init__(self, parent, directories=False):
            QFileDialog.__init__(self)
            self.setWindowTitle("PURR: Add Data Products")
            self.dirlist = None
            # add filters
            # self._dirfilter = "Any directories (*)"
            # if hasattr(self,'setNameFilters'):  # qt4.4+
            #  self.setNameFilters(["Any files (*)",self._dirfilter])
            # else:
            #  self.setFilters(["Any files (*)",self._dirfilter])
            # connect signals (translates into a SIGNAL with string arguments)
            self.connect(self, SIGNAL("filesSelected(const QStringList&)"), self._filesSelected)
            #      self.connect(self,SIGNAL("fileSelected(const QString&)"),self._fileSelected)
            self.connect(self, SIGNAL("currentChanged(const QString&)"), self._fileHighlighted)
            # self.connect(self,SIGNAL("filterSelected(const QString&)"),self._filterSelected)
            # resize selves
            self.config_name = "add-file-dialog"
            width = Config.getint('%s-width' % self.config_name, 768)
            height = Config.getint('%s-height' % self.config_name, 512)
            self.resize(QSize(width, height))

        def resizeEvent(self, ev):
            QFileDialog.resizeEvent(self, ev)
            sz = ev.size()
            Config.set('%s-width' % self.config_name, sz.width())
            Config.set('%s-height' % self.config_name, sz.height())

        def show(self):
            QFileDialog.show(self)
            self._file = None

        #    def done (self,code):
        #      """Workaround for QFileDialog bug: if DirctoryOnly mode, it doesn't actually emit any
        #      fileSelected() when OK is pressed. So we catch the selected dir via fileHighlighted()
        #      below, and report it here."""
        #      if code == 1 and self.mode() == QFileDialog.DirectoryOnly and self._file:
        #        self.emit(SIGNAL("filesSelected"),self._file)
        #      QFileDialog.done(self,code)

        def _filterSelected(self, file_filter):
            if str(file_filter) == self._dirfilter:
                self.setFileMode(QFileDialog.Directory)
                self.setOption(QFileDialog.ShowDirsOnly)
            else:
                self.setFileMode(QFileDialog.ExistingFiles)
                self.setOption(0)

        def setDirList(self, dirlist):
            if dirlist is not self.dirlist:
                self.dirlist = dirlist
                if dirlist and hasattr(self, 'setSidebarUrls'):
                    self.setSidebarUrls([QUrl.fromLocalFile(path) for path in dirlist])

        def _filesSelected(self, filelist):
            self.emit(SIGNAL("filesSelected"), *sorted(map(str, filelist)))

        def _fileSelected(self, file):
            self.emit(SIGNAL("filesSelected"), str(file))

        def _fileHighlighted(self, file):
            self._file = str(file)

    def _showAddFileDialog(self, add_dir=False):
        # create dialog when first called
        dialog = self._add_dp_dialog
        if not dialog:
            self._add_dp_dialog = dialog = self.AddDataProductDialog(self)
            self.connect(dialog, SIGNAL("filesSelected"), self, SIGNAL("filesSelected"))
            dialog.setDirectory(self._default_dirs[0] if self._default_dirs else ".")
        # set mode
        if add_dir:
            dialog.setFileMode(QFileDialog.Directory)
        else:
            dialog.setFileMode(QFileDialog.ExistingFiles)
        # add quick-jump combobox
        dialog.setDirList(list(self._default_dirs) + [os.path.expanduser("~")])
        dialog.setDirectory(dialog.directory());  # hope this is the same as rereadDir() in qt3
        dialog.show()
        dialog.raise_()

    def _showAddDirDialog(self):
        return self._showAddFileDialog(add_dir=True)

    def _titleChanged(self, *dum):
        self._title_changed = True
        self.setUpdated()

    def _commentChanged(self, *dum):
        self._comment_changed = self._comment_edited = True
        self.setUpdated()

    def suggestTitle(self, title):
        """Suggests a title for the entry.
        If title has been manually edited, suggestion is ignored."""
        if not self._title_changed or not str(self.wtitle.text()):
            self.wtitle.setText(title)
            self._title_changed = False

    def addComment(self, comment):
        # get current comment text, if nothing was changed at all, use empty text
        if self._comment_changed:
            cur_comment = str(self.comment_doc.toPlainText())
            cpos = self.wcomment.textCursor()
        else:
            cur_comment = ""
            cpos = None
        # no current comments? Replace
        if not cur_comment:
            cur_comment = comment
        # else does the comment still end with the last auto-added string? Simply
        # append our comment then
        elif self._last_auto_comment and cur_comment.endswith(self._last_auto_comment):
            cur_comment += comment
        # else user must've edited the end of the comment. Start a new paragraph and append our
        # comment
        else:
            cur_comment += "\n\n" + comment
        self._comment_edited = False
        self._last_auto_comment = comment
        # update widget
        self.comment_doc.setPlainText(cur_comment)
        if cpos:
            self.wcomment.setTextCursor(cpos)

    def countRemovedDataProducts(self):
        """Returns number of DPs marked for removal"""
        return len([item for item, dp in self.wdplv.getItemDPList() if dp.policy == "remove"])

    def resolveFilenameConflicts(self, dialog=True):
        """Goes through list of DPs to make sure that their destination names
        do not clash. Applies new names. Returns True if some conflicts were resolved.
        If dialog is True, shows confirrmation dialog."""
        resolved = self.wdplv.resolveFilenameConflicts()
        if resolved and dialog:
            QMessageBox.warning(self, "Filename conflicts", """<P>
        <NOBR>PURR has found duplicate destination filenames among your data products.</NOBR>
        This is not allowed, so some filenames have been adjusted to avoid name clashes.
        Please review the changes before saving this entry.
        </P>""",
                                QMessageBox.Ok, 0)
        return resolved

    def updateEntry(self):
        """Updates entry object with current content of dialog.
        In new entry mode (setEntry() not called, so self.entry=None), creates new entry object.
        In old entry mode (setEntry() called), updates and saves old entry object.
        """
        # form up new entry
        title = str(self.wtitle.text())
        comment = str(self.comment_doc.toPlainText())
        # process comment string -- eliminate single newlines, make double-newlines into separate paragraphs
        # exception are paragraphs that start with "#LOG:", these get special treatment and the single newlines are
        # left intact
        pars = []
        for paragraph in comment.split("\n\n"):
            if paragraph.startswith("LOG:"):
                pars.append(paragraph.replace("\n", "<BR>"))
            else:
                pars.append(paragraph.replace("\n", " "))
        comment = "\n".join(pars)
        # go through data products and decide what to do with each one
        busy = Purr.BusyIndicator()
        # get list of DPs
        dps, updated = self.wdplv.buildDPList()
        # emit signal for all newly-created DPs
        for dp in dps:
            if not dp.archived:
                self.emit(SIGNAL("creatingDataProduct"), dp.sourcepath)
        # update or return new entry
        if self.entry:
            self.entry.update(title=title, comment=comment, dps=dps)
            self.entry.save(refresh_index=updated)
            return self.entry
        else:
            return Purr.LogEntry(time.time(), title, comment, dps)

    def updateIgnoredEntry(self):
        """Updates an ignore-entry object with current content of dialog, by
        marking all data products for ignore."""
        # collect new DPs from items
        dps = []
        for item, dp in self.wdplv.getItemDPList():
            if dp and not dp.archived:  # None means a new DP
                # set all policies to ignore, unless already set to banish
                if dp.policy != "banish":
                    self.wdplv.setItemPolicy(item, "ignore")
                    dp.set_policy("ignore")
                dps.append(dp)
        # return new entry
        return Purr.LogEntry(time.time(), dps=dps, ignore=True)

    def setEntry(self, entry=None):
        """Populates the dialog with contents of an existing entry."""
        busy = Purr.BusyIndicator()
        self.entry = entry
        self.setEntryTitle(entry.title)
        self.setEntryComment(entry.comment.replace("\n", "\n\n").replace("<BR>", "\n"))
        self.wdplv.clear()
        self.wdplv.fillDataProducts(entry.dps)
        self.setTimestamp(entry.timestamp)
        self.updated = False

    def addDataProducts(self, dps):
        """Adds data products to dialog.
          dps is a list of DP objects
        """
        return self.wdplv.addDataProducts(dps)

    def dropDataProducts(self, *pathnames):
        """Drops new (i.e. non-archived) DP items matching the given pathnames."""
        return self.wdplv.dropDataProducts(*pathnames)

    def entryTitle(self):
        return self.wtitle.text()

    def setEntryTitle(self, title, select=True):
        self.wtitle.setText(title)
        if select:
            self.wtitle.selectAll()

    def setEntryComment(self, comment, select=True):
        self.comment_doc.setPlainText(comment)
        if select:
            self.wcomment.selectAll()

    def setTimestamp(self, timestamp):
        self._timestamp = timestamp
        txt = time.strftime("%x %X", time.localtime(timestamp))
        self.wtimestamp.setText(txt)
        self.wtimestamp.show()


class NewLogEntryDialog(QDialog):
    # class DialogTip (QToolTip):
    # def __init__ (self,parent):
    # QToolTip.__init__(self,parent)
    # self.parent = parent
    # def maybeTip (self,pos):
    # parent = self.parent
    # if parent._has_tip:
    # rect = QRect(pos.x()-20,pos.y()-20,40,40)
    # self.tip(rect,parent._has_tip)
    # parent._has_tip = None

    def __init__(self, parent, *args):
        QDialog.__init__(self, parent, *args)
        self.setWindowTitle("Adding Log Entry")
        self.setWindowIcon(pixmaps.purr_logo.icon())
        self.setModal(False)
        ## create pop-up tip
        # self._has_tip = None
        # self._dialog_tip = self.DialogTip(self)
        # create editor
        lo = QVBoxLayout(self)
        lo.setMargin(5)
        # lo.setResizeMode(QLayout.Minimum)
        self.editor = LogEntryEditor(self)
        self.dropDataProducts = self.editor.dropDataProducts
        self.connect(self.editor, SIGNAL("filesSelected"), self, SIGNAL("filesSelected"))
        # connect draggedAwayFiles() signal so that files dragged away are removed from
        # the list
        self.connect(self.editor, SIGNAL("draggedAwayFiles"), self.editor.dropDataProducts)
        lo.addWidget(self.editor)
        lo.addSpacing(5)
        # create button bar
        btnfr = QFrame(self)
        btnfr.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        # btnfr.setMargin(5)
        lo.addWidget(btnfr)
        lo.addSpacing(5)
        btnfr_lo = QHBoxLayout(btnfr)
        btnfr_lo.setMargin(0)
        newbtn = QPushButton(pixmaps.filesave.icon(), "Add new entry", btnfr)
        newbtn.setToolTip("""<P>Saves new log entry, along with any data products not marked
      as "ignore" or "banish".<P>""")
        self.ignorebtn = QPushButton(pixmaps.red_round_cross.icon(), "Ignore all", btnfr)
        self.ignorebtn.setEnabled(False)
        self.ignorebtn.setToolTip("""<P>Tells PURR to ignore all listed data products. PURR
      will not pounce on these files again until they have been modified.<P>""")
        cancelbtn = QPushButton(pixmaps.grey_round_cross.icon(), "Hide", btnfr)
        cancelbtn.setToolTip("""<P>Hides this dialog.<P>""")
        QObject.connect(newbtn, SIGNAL("clicked()"), self.addNewEntry)
        QObject.connect(self.ignorebtn, SIGNAL("clicked()"), self.ignoreAllDataProducts)
        QObject.connect(cancelbtn, SIGNAL("clicked()"), self.hide)
        for btn in (newbtn, self.ignorebtn, cancelbtn):
            btn.setAutoDefault(False)
        btnfr_lo.setMargin(0)
        btnfr_lo.addWidget(newbtn, 2)
        btnfr_lo.addStretch(1)
        btnfr_lo.addWidget(self.ignorebtn, 2)
        btnfr_lo.addStretch(1)
        btnfr_lo.addWidget(cancelbtn, 2)
        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        # resize selves
        self.setMinimumSize(256, 512)
        width = Config.getint('entry-editor-width', 512)
        height = Config.getint('entry-editor-height', 512)
        self.resize(QSize(width, height))

    def resizeEvent(self, ev):
        QDialog.resizeEvent(self, ev)
        sz = ev.size()
        Config.set('entry-editor-width', sz.width())
        Config.set('entry-editor-height', sz.height())

    def showEvent(self, ev):
        QDialog.showEvent(self, ev)
        self.emit(SIGNAL("shown"), True)

    def hideEvent(self, ev):
        QDialog.hideEvent(self, ev)
        self.emit(SIGNAL("shown"), False)

    def reset(self):
        self.editor.reset()

    def addDataProducts(self, dps):
        updated = self.editor.addDataProducts(dps)
        self.ignorebtn.setEnabled(True)
        return updated

    def suggestTitle(self, title):
        self.editor.suggestTitle(title)

    def addComment(self, comment):
        self.editor.addComment(comment)

    def setDefaultDirs(self, *dirs):
        self.editor.setDefaultDirs(*dirs)

    def ignoreAllDataProducts(self):
        # confirm with user
        if QMessageBox.question(self, "Ignoring all data products", """<P><NOBR>Do you really
          want to ignore all data products</NOBR> listed here? PURR will ignore these
          files until they have been modified again.""",
                                QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
            return
        # add ingore entry, which will store info on all ignored data products
        entry = self.editor.updateIgnoredEntry()
        self.emit(SIGNAL("newLogEntry"), entry)
        self.editor.resetDPs()
        self.hide()

    def addNewEntry(self):
        # if some naming conflicts have been resolved, return -- user will need to re-save
        if self.editor.resolveFilenameConflicts():
            return
        # confirm with user
        if QMessageBox.question(self, "Adding new entry", """<P><NOBR>Do you really want to add
          a new log entry</NOBR> titled "%s"?""" % (self.editor.entryTitle()),
                                QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
            return
        # add entry
        entry = self.editor.updateEntry()
        self.emit(SIGNAL("newLogEntry"), entry)
        self.editor.reset()
        self.hide()


class ExistingLogEntryDialog(QDialog):
    def __init__(self, parent, *args):
        QDialog.__init__(self, parent, *args)
        self.setModal(False)
        # make stack for viewer and editor components
        lo = QVBoxLayout(self)
        lo.setMargin(5)
        self.wstack = QStackedWidget(self)
        lo.addWidget(self.wstack)
        self.wstack.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        # make editor panel
        self.editor_panel = QWidget(self.wstack)
        self.editor_panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.wstack.addWidget(self.editor_panel)
        lo = QVBoxLayout(self.editor_panel)
        lo.setMargin(0)
        # create editor
        self.editor = LogEntryEditor(self.editor_panel)
        self.connect(self.editor, SIGNAL("updated"), self._entryUpdated)
        self.connect(self.editor, SIGNAL("filesSelected"), self, SIGNAL("filesSelected"))
        self.connect(self.editor, SIGNAL("creatingDataProduct"),
                     self, SIGNAL("creatingDataProduct"))
        lo.addWidget(self.editor)
        self.dropDataProducts = self.editor.dropDataProducts
        # create button bar for editor
        btnfr = QFrame(self.editor_panel)
        btnfr.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        # btnfr.setMargin(5)
        lo.addWidget(btnfr)
        lo.addSpacing(5)
        btnfr_lo = QHBoxLayout(btnfr)
        self.wsave = QPushButton(pixmaps.filesave.icon(), "Save", btnfr)
        QObject.connect(self.wsave, SIGNAL("clicked()"), self._saveEntry)
        cancelbtn = QPushButton(pixmaps.grey_round_cross.icon(), "Cancel", btnfr)
        QObject.connect(cancelbtn, SIGNAL("clicked()"), self._cancelEntry)
        for btn in (self.wsave, cancelbtn):
            btn.setAutoDefault(False)
        btnfr_lo.setMargin(0)
        btnfr_lo.addWidget(self.wsave, 1)
        btnfr_lo.addStretch(1)
        btnfr_lo.addWidget(cancelbtn, 1)

        # create viewer panel
        self.viewer_panel = QWidget(self.wstack)
        self.viewer_panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.wstack.addWidget(self.viewer_panel)
        lo = QVBoxLayout(self.viewer_panel)
        lo.setMargin(0)
        label = QLabel("""<P>Below is an HTML rendering of your log entry. Note that this is
      only a bare-bones viewer, so only some links will work. In particular,
      you may click on a link associated with a saved data product to get a menu of related actions. To edit this entry, click the Edit button below.
      </P>""", self.viewer_panel)
        label.setWordWrap(True)
        label.setMargin(5)
        lo.addWidget(label)
        self.viewer = self.EntryTextBrowser(self.viewer_panel)
        self.viewer.setOpenLinks(False)
        self.viewer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self._viewer_source = None
        QObject.connect(self.viewer, SIGNAL("anchorClicked(const QUrl &)"), self._urlClicked)
        lo.addWidget(self.viewer)
        lo.addSpacing(5)
        # create button bar
        btnfr = QFrame(self.viewer_panel)
        btnfr.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)
        # btnfr.setMargin(5)
        lo.addWidget(btnfr)
        lo.addSpacing(5)
        btnfr_lo = QHBoxLayout(btnfr)
        btnfr_lo.setMargin(0)
        btn = self.wprev = QPushButton(pixmaps.previous.icon(), "Previous", btnfr)
        QObject.connect(btn, SIGNAL("clicked()"), self, SIGNAL("previous()"))
        btnfr_lo.addWidget(btn, 1)
        btnfr_lo.addSpacing(5)
        btn = self.wnext = QPushButton(pixmaps.next.icon(), "Next", btnfr)
        QObject.connect(btn, SIGNAL("clicked()"), self, SIGNAL("next()"))
        btnfr_lo.addWidget(btn, 1)
        btnfr_lo.addSpacing(5)
        btn = self.wedit = QPushButton(pixmaps.edit.icon(), "Edit", btnfr)
        QObject.connect(btn, SIGNAL("clicked()"), self._editEntry)
        btnfr_lo.addWidget(btn, 1)
        btnfr_lo.addStretch(1)
        btn = self.wclose = QPushButton(pixmaps.grey_round_cross.icon(), "Close", btnfr)
        QObject.connect(btn, SIGNAL("clicked()"), self.hide)
        btnfr_lo.addWidget(btn, 1)
        # create popup menu for data products
        self._dp_menu = menu = QMenu(self)
        self._dp_menu_title = QLabel()
        self._dp_menu_title.setMargin(5)
        self._dp_menu_title_wa = wa = QWidgetAction(self)
        wa.setDefaultWidget(self._dp_menu_title)
        menu.addAction(wa)
        menu.addSeparator()
        menu.addAction(pixmaps.editcopy.icon(), "Restore file(s) from archived copy", self._restoreDPFromArchive)
        menu.addAction(pixmaps.editpaste.icon(), "Copy pathname of archived copy to clipboard", self._copyDPToClipboard)
        menu.addAction("Drag-and-drop archived copy", self._startDPDrag)
        self._dp_menu_on = None

        # resize selves
        width = Config.getint('entry-viewer-width', 512)
        height = Config.getint('entry-viewer-height', 512)
        self.resize(QSize(width, height))
        # other init
        self.entry = None
        self.updated = False

    class EntryTextBrowser(QTextBrowser):
        def mouseReleaseEvent(self, ev):
            self._mouse_release_pos = ev.globalPos()
            return QTextBrowser.mouseReleaseEvent(self, ev)

        def getLastMouseRelease(self):
            try:
                return self._mouse_release_pos
            except AttributeError:
                return self.mapToGlobal(QPoint(0, 0))

    def resizeEvent(self, ev):
        QDialog.resizeEvent(self, ev)
        sz = ev.size()
        Config.set('entry-viewer-width', sz.width())
        Config.set('entry-viewer-height', sz.height())

    def viewEntry(self, entry, prev=None, next=None):
        # if editing previous entry, ask for confirmation
        if self.updated:
            self.show()
            self._saveEntry()
        busy = Purr.BusyIndicator()
        self.entry = entry
        self.updated = False
        self.setWindowTitle(entry.title)
        self._viewer_source = QUrl.fromLocalFile(self.entry.index_file)
        self.viewer.setSource(self._viewer_source)
        self.viewer.reload()
        # self._prev_path = prev and prev.index_file
        # self._next_path = next and next.index_file
        self.wprev.setEnabled(bool(prev))
        self.wnext.setEnabled(bool(next))
        self.wstack.setCurrentWidget(self.viewer_panel)

    def _urlClicked(self, url):
        # if self._viewer_source:
        # self.viewer.setSource(self._viewer_source)
        # see if we clicked on a URL for a data product, display URL for it if so
        path = str(url.path())
        if path:
            # see if we clicked on a URL for a data product, display URL for it if so
            for dp in self.entry.dps:
                if os.path.samefile(dp.fullpath, path) or os.path.exists(dp.subproduct_dir()) and os.path.samefile(
                        dp.subproduct_dir(), os.path.dirname(path)):
                    self._dp_menu_title.setText(os.path.basename(dp.filename))
                    self._dp_menu_on = dp
                    self._dp_menu.exec_(self.viewer.getLastMouseRelease())
                    return
            # else emit signal
            self.emit(SIGNAL("viewPath"), path)

    def _copyDPToClipboard(self):
        """Callback for item menu."""
        dp = self._dp_menu_on
        if dp and dp.archived:
            path = dp.fullpath.replace(" ", "\\ ")
            QApplication.clipboard().setText(path, QClipboard.Clipboard)
            QApplication.clipboard().setText(path, QClipboard.Selection)

    def _restoreDPFromArchive(self):
        """Callback for item menu."""
        dp = self._dp_menu_on
        if dp and dp.archived:
            dp.restore_from_archive(parent=self)

    def _startDPDrag(self):
        """Callback for item menu."""
        dp = self._dp_menu_on
        if dp and dp.archived:
            drag = QDrag(self)
            mimedata = QMimeData()
            mimedata.setUrls([QUrl.fromLocalFile(dp.fullpath)])
            drag.setMimeData(mimedata)
            drag.exec_(Qt.CopyAction | Qt.LinkAction)

    def setDefaultDirs(self, *dirs):
        self.editor.setDefaultDirs(*dirs)

    def addDataProducts(self, dps):
        self.editor.addDataProducts(dps)
        self._entryUpdated()

    def _editEntry(self):
        self.setWindowTitle("Editing entry")
        self.editor.setEntry(self.entry)
        self.updated = False
        self.wsave.setEnabled(False)
        self.wstack.setCurrentWidget(self.editor_panel)

    def _saveEntry(self):
        # if some naming conflicts have been resolved, return -- user will need to re-save
        if self.editor.resolveFilenameConflicts():
            return
        # ask for confirmation
        nremove = self.editor.countRemovedDataProducts()
        msg = "<P><nobr>Save changes to this log entry?</nobr></P>"
        if nremove:
            msg += "<P>%d archived data product(s) will be removed.</P>" % nremove
        if QMessageBox.question(self, "Saving entry", msg,
                                QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
            return
        busy = Purr.BusyIndicator()
        self.editor.updateEntry()
        self.updated = False
        self.setWindowTitle(self.entry.title)
        self.viewer.reload();  # setSource(QUrl.fromLocalFile(self.entry.index_file))
        self.wstack.setCurrentWidget(self.viewer_panel)
        # emit signal to regenerate log
        self.emit(SIGNAL("entryChanged"), self.entry)

    def _cancelEntry(self):
        if self.updated and QMessageBox.question(self, "Abandoning changes",
                                                 "Abandon changes to this log entry?",
                                                 QMessageBox.Yes, QMessageBox.No) != QMessageBox.Yes:
            return
        self.setWindowTitle(self.entry.title)
        self.wstack.setCurrentWidget(self.viewer_panel)

    def _entryUpdated(self):
        self.updated = True
        self.wsave.setEnabled(True)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import os.path
import re
import sys
from types import *
from . import chart_object

from .pychart_types import *

oldstdout = sys.stdout
if os.path.exists("/dev/null"):
    sys.stdout = open("/dev/null", "w")

modules = {}
values = []

sys.stdout = oldstdout
g = globals()
for mod in list(g.keys()):
    val = g[mod]
    if type(val) == ModuleType:
        dic = {}
        for name in list(val.__dict__.keys()):
            v = val.__dict__[name]
            if name[0] != '_':
                values.append((v, mod + "." + name))
            if type(v) == type and issubclass(v, chart_object.T):
                dic[name] = v
        modules[mod] = dic


def stringify_type(t):
    s = str(t)
    if t == AnyType:
        return "any"
    if t == ShadowType:
        return "(xoff,yoff,fill)"
    elif re.search("NumberType", s):
        return "number"
    elif re.search("UnitType", s):
        return 'length in points (\\xref{unit})'
    elif re.search("CoordType", s):
        return "(x,y) or None"
    elif re.search("CoordSystemType", s):
        return "['linear'|'log'|'category']"
    elif re.search("TextAlignType", s):
        return "['R'|'L'|'C'|None]"
    elif re.search("FormatType", s):
        return "printf format string"
    elif re.search("IntervalType", s):
        return "number or function"

    mo = re.match("<type '([^']+)'>", s)
    if mo:
        return mo.group(1)
    mo = re.match("<class 'pychart\.([^']+)'>", s)
    if mo:
        return mo.group(1)
    mo = re.match("<class '([^']+)'>", s)
    if mo:
        return mo.group(1)
    mo = re.match("pychart\\.(.*)", s)
    if mo:
        return mo.group(1)
    return s


def stringify_value(val):
    t = type(val)
    if t == StringType:
        return '"' + val + '"'
    if t == bool:
        if val:
            return "True"
        else:
            return "False"

    if t in (IntType, LongType, FloatType):
        return str(val)
    if val == None:
        return "None"
    if type(val) == ListType:
        return list(map(stringify_value, val))
    for pair in values:
        if pair[0] == val:
            return pair[1]
    return str(val)


def break_string(name):
    max_len = 10
    if len(name) < max_len:
        return name

    name = re.sub("(\\d\\d)([^\\d])", "\\1-\n\\2", name)
    name = re.sub("black(.)", "black-\n\\1", name)

    elems = name.split("\n")
    while 1:
        broken = 0
        for i in range(len(elems)):
            elem = elems[i]
            if len(elem) < max_len:
                continue
            broken = 1
            elem1 = elem[0:len(elem) / 2]
            elem2 = elem[len(elem) / 2:]
            elems[i:i + 1] = [elem1, elem2]
            break
        if not broken:
            break
    name = "\n".join(elems)
    return name



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import sys

from . import gs_frontend
from . import theme


class T(gs_frontend.T):
    def close(self):
        gs_frontend.T.close(self)
        if self.__output_lines == []:
            return

        if theme.use_color:
            gs_args = "-sDEVICE=png256 -dTextAlphaBits=4 -q -dNOPAUSE"  # PDS
        else:
            gs_args = "-sDEVICE=pnggray -dTextAlphaBits=4 -q -dNOPAUSE"  # PDS

        temp_fname = None  # the temporary file desc.
        out_fd = None  # the final destination. 

        if self.__out_fname and isinstance(self.__out_fname, str):
            gs_args += " -sOutputFile=%s" % self.__out_fname
        else:
            if not self.__out_fname:
                out_fd = sys.stdout
            else:
                if not hasattr(self.__out_fname, "write"):
                    raise Exception("Expecting either a filename or a file-like object, but got %s" % self.__out_fname)
                out_fd = self.__out_fname
            import tempfile
            temp_fname = tempfile.mktemp()
            gs_args += " -sOutputFile=%s" % temp_fname
        self.start_gs(gs_args)
        self.close_gs()

        if temp_fname:
            temp_fd = open(temp_fname, 'rb')
            out_fd.write(temp_fd.read())
            temp_fd.close()



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
import atexit
import re
import sys

from . import basecanvas
from . import pdfcanvas
from . import pngcanvas
from . import pscanvas
from . import svgcanvas
from . import theme
from . import x11canvas

invalid_coord = -999999
_oldexitfunc = None

T = basecanvas.T


def init(fname=None, format=None):
    """This is a "factory" procedure that creates a new canvas.T object.
    Both parameters, <fname> and
    <format>, are optional. Parameter <fname> specifies either the output
    file name or a file object. Parameter <format>, if specified, defines the
    file's format. Its value must be one of "ps", "pdf", "svg", "x11", or
    "png".

    When <fname> is omitted or is None, the output is sent to standard
    output. When <format> is omitted, it is guessed from the <fname>'s
    suffix; failing that, "ps" is selected."""

    fname = fname or theme.output_file
    format = format or theme.output_format

    if format == None:
        if not isinstance(fname, str):
            format = "ps"
        elif re.search("pdf$", fname):
            format = "pdf"
        elif re.search("png$", fname):
            format = "png"
        elif re.search("svg$", fname):
            format = "svg"
        else:
            format = "ps"

    if format == "ps":
        can = pscanvas.T(fname)
    elif format == "png":
        can = pngcanvas.T(fname)
    elif format == "x11":
        can = x11canvas.T(fname)
    elif format == "svg":
        can = svgcanvas.T(fname)
    elif format == "pdf-uncompressed":
        can = pdfcanvas.T(fname, False)
    else:
        can = pdfcanvas.T(fname, theme.compress_output)
    return can


def default_canvas():
    if len(basecanvas.active_canvases) > 0:
        return basecanvas.active_canvases[0]
    else:
        return init(None)


def _exit():
    global _oldexitfunc, _active_canvases

    for can in basecanvas.active_canvases[:]:
        can.close()

    if _oldexitfunc:
        foo = _oldexitfunc
        _oldexitfunc = None
        foo()


#
# The following procedures are there just for backward compatibility.
# 
def line(style, x1, y1, x2, y2):
    default_canvas().line(style, x1, y1, x2, y2)


def curve(style, points):
    default_canvas().curve(style, points)


def clip(x1, y1, x2, y2):
    default_canvas().clip(x1, y1, x2, y2)


def endclip():
    default_canvas().endclip()


def clip_polygon(points):
    default_canvas().clip_polygon(points)


def clip_ellipsis(x, y, radius, ratio=1.0):
    default_canvas().clip_ellipsis(x, y, radius, ratio)


def ellipsis(line_style, pattern, x, y, radius, ratio=1.0,
             start_angle=0, end_angle=360, shadow=None):
    default_canvas().ellipsis(line_style, pattern, x, y, radius, ratio,
                              start_angle, end_angle, shadow)


def rectangle(edge_style, pat, x1, y1, x2, y2, shadow=None):
    default_canvas().rectangle(edge_style, pat, x1, y1, x2, y2, shadow)


def polygon(edge_style, pat, points, shadow=None):
    default_canvas().polygon(edge_style, pat, points, shadow=None)


def close():
    default_canvas().close()


def round_rectangle(style, fill, x1, y1, x2, y2, radius, shadow=None):
    default_canvas().round_rectangle(style, fill, x1, y1, x2, y2, radius, shadow)


def show(x, y, str):
    default_canvas().show(x, y, str)


atexit.register(_exit)

# theme.add_reinitialization_hook(lambda: init(None))



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import canvas
from . import chart_object
from . import color
from . import fill_style
from . import font
from . import legend
from . import line_style
from . import pychart_util
from .pychart_types import *


class T(chart_object.T):
    """Plots sector diagram which can be superimposed on one another.
    Sector diagrams are also known as wind roses"""
    keys = {
        "start_angle": (NumberType, 90, ""),  # top of chart (north)
        "center": (CoordType, None, ""),
        "base_radius": (NumberType, None, ""),
        "line_style": (line_style.T, line_style.T(color=color.black, width=0.3), ""),
        "fill_styles": (list, fill_style.standards.list()[:],  # type: ignore
                        """The fill style of each item. The length of the
                        list should be equal to the length of the data.
                        """),
        "sector_centred": (int, 1,
                           """Bool indicating whether the sectors should be centred on each sector_width(e.g. on 0)"""),
        "dir_offset": (UnitType, None,
                       """The distance between the directions and the outermost circle. Defaults fine for most cases"""),
        "data": (AnyType, None, pychart_util.data_desc),
        "label_col": (int, 0,
                      """The column, within "data", from which the labels of items are retrieved."""),
        "data_col": (int, 1,
                     """ The column, within "data", from which the data values are retrieved."""),
        "dir_line_style": (line_style.T, None, ""),
        "dir_fill_style": (fill_style.T, fill_style.default, ""),
        "shadow": (ShadowType, None, pychart_util.shadow_desc),
        "sector_width": (int, None, ""),  # automatically generated
    }

    def __init__(self, colour=True, **args):
        chart_object.T.init(self, args)
        if colour:
            # the theme.color flag does not seem to affect the fill_style.standards,
            # besides, I want the first two colors to resemble those of gnuplot's postscript terminal
            self.fill_styles = [fill_style.Plain(bgcolor=color.red),
                                fill_style.Plain(bgcolor=color.green),
                                fill_style.Plain(bgcolor=color.blue),
                                fill_style.Plain(bgcolor=color.magenta)]

    def check_integrity(self):
        nSectors = len(self.data[0][self.data_col])
        if (360 % nSectors != 0):
            raise Exception('Length of dataset ' + str(nSectors) + ' not a divisor of 360 degrees!')
        for dataset in self.data:
            length = len(dataset[self.data_col])
            if length != nSectors:
                raise Exception('Lengths of datasets given is different!')
            for val in dataset[self.data_col]:
                if (val < 0) | (val > 1):
                    raise Exception('Data value ' + str(val) + ' not between 0 and 1!')
        self.sector_width = 360 / nSectors
        chart_object.T.check_integrity(self)

    def get_data_range(self, which):
        return (0, 1)

    def get_legend_entry(self):
        legends = []
        i = 0
        for dataset in self.data:
            fill = self.fill_styles[i]
            i = (i + 1) % len(self.fill_styles)
            legends.append(legend.Entry(line_style=self.line_style,
                                        fill_style=fill,
                                        label=dataset[self.label_col]))
        return legends

    def draw(self, ar, can):
        center = self.center
        if not center:
            center = (ar.loc[0] + ar.size[0] / 2.0,
                      ar.loc[1] + ar.size[1] / 2.0)
        base_radius = self.base_radius  # the maximum radius of a wedge
        if not base_radius:
            base_radius = min(ar.size[0] / 2.0, ar.size[1] / 2.0)  # * 0.8

        sector_decrement = 1. / (len(
            self.data) * 2) * self.sector_width  # each following sector diagram will have its sector width decremented by half this amount (in degrees)
        i = 0
        for dataset in self.data:
            cur_angle = self.start_angle
            if self.sector_centred:
                cur_angle -= self.sector_width / 2.
            fill = self.fill_styles[i]
            x_center = center[0]
            y_center = center[1]

            if not i:  # draw directions around sector diagram once off
                dir_offset = base_radius + (self.dir_offset or base_radius * 0.04)
                directions = ['N', 'E', 'S', 'W']
                angle = self.start_angle

                can.ellipsis(line_style.T(color=color.black, width=0.3, dash=line_style.dash1), None,
                             x_center, y_center, base_radius, 1,
                             0, 360)  #

                for d in directions:
                    x_label, y_label = pychart_util.rotate(dir_offset, 0, angle)  # coords for bottom left corner of box
                    tw = font.text_width(d)
                    half = 1 / 3.  # normal arithmetic does not seem to apply to these text_box objects...
                    if (angle == 0):  # east
                        y_label -= font.text_height(d)[0] * half  # move down half
                    elif (angle == -180):  # west
                        y_label -= font.text_height(d)[0] * half  # move down half
                        x_label -= font.text_width(d)  # move left full
                    elif (angle == 90):  # north
                        x_label -= font.text_height(d)[0] * half  # move left half
                    elif (angle == -90):  # south
                        y_label -= font.text_height(d)[
                                       0] * .8  # move down (couldn't figure out how to set this dynamically so I fudged...)
                        x_label -= font.text_height(d)[0] * half  # move left half
                    canvas.show(x_label + x_center, y_label + y_center, d)
                    angle -= 360 / len(directions)

            for val in dataset[self.data_col]:  # now draw the sectors
                radius = base_radius * val  # scale the radius
                start = cur_angle - self.sector_width + i * sector_decrement
                stop = cur_angle - i * sector_decrement  # these may seem confusing, but remember that we need to go counterclockwise

                can.ellipsis(self.line_style, fill,
                             x_center, y_center, radius, 1, start, stop, self.shadow)
                cur_angle = (
                                        cur_angle - self.sector_width) % 360  # we want to go in anticlockwise direction (North, West, South, etc. as in meteorology)
            i = (i + 1) % len(self.fill_styles)



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import fill_style
from . import line_style
from . import object_set
from . import tick_mark_doc
from . import pychart_util
from .pychart_types import *

_keys = {
    "line_style": (line_style.T, line_style.default, "The line style of the tick mark."),
    "fill_style": (fill_style.T, fill_style.white, "The fill style."),
    "size": (UnitType, 5, "Size of the tick mark."),
}


class T(chart_object.T):
    __doc__ = tick_mark_doc.doc
    keys = _keys


##AUTOMATICALLY GENERATED

##END AUTOMATICALLY GENERATED

class Circle(T):
    """Draws a circle. """

    def draw(self, can, x, y):
        can.ellipsis(self.line_style, self.fill_style, x, y,
                     self.size / 2.0, 1)


class Square(T):
    """Draws a square."""

    def draw(self, can, x, y):
        # move to the bottom-left corner
        x = x - self.size / 2.0
        y = y - self.size / 2.0
        can.rectangle(self.line_style, self.fill_style,
                      x, y, x + self.size, y + self.size)


class Triangle(T):
    """Draws a triangle pointing up."""

    def draw(self, can, x, y):
        can.polygon(self.line_style, self.fill_style,
                    ((x - self.size / 1.6, y - self.size / 2.0),
                     (x + self.size / 1.6, y - self.size / 2.0),
                     (x, y + self.size / 2.0)))


class DownTriangle(T):
    """Draws a triangle pointing down."""

    def draw(self, can, x, y):
        can.polygon(self.line_style, self.fill_style,
                    ((x, y - self.size / 2.0),
                     (x - self.size / 1.6, y + self.size / 2.0),
                     (x + self.size / 1.6, y + self.size / 2.0)))


class X(T):
    """Draw a "X"-shaped tick mark. Attribute "fill-style" is ignored."""
    keys = pychart_util.union_dict(T.keys,
                                   {"line_style": (line_style.T,
                                                   line_style.T(width=0.7),
                                                   "The line style of the tick mark")})

    def draw(self, can, x, y):
        # move to the bottom-left corner
        x = x - self.size / 2.0
        y = y - self.size / 2.0
        can.line(self.line_style, x, y, x + self.size, y + self.size)
        can.line(self.line_style, x + self.size, y, x, y + self.size)


class Plus(T):
    """Draw a "+"-shaped tick mark. Attribute "fill-style" is ignored."""
    keys = pychart_util.union_dict(T.keys,
                                   {"line_style": (line_style.T,
                                                   line_style.T(width=1),
                                                   "The line style of the tick mark.")})

    def draw(self, can, x, y):
        # move to the bottom-left corner
        can.line(self.line_style, x - self.size / 1.4, y, x + self.size / 1.4, y)
        can.line(self.line_style, x, y - self.size / 1.4, x, y + self.size / 1.4)


class Diamond(T):
    """Draw a square rotated at 45 degrees."""

    def draw(self, can, x, y):
        # move to the bottom-left corner
        can.polygon(self.line_style, self.fill_style,
                    ((x - self.size / 1.4, y), (x, y + self.size / 1.4),
                     (x + self.size / 1.4, y), (x, y - self.size / 1.4)))


class Star(T):
    """Draw a "*". Attribute "fill-style" is ignored."""
    keys = pychart_util.union_dict(T.keys,
                                   {"line_style": (line_style.T,
                                                   line_style.T(width=1),
                                                   "The line style of the tick mark.")})

    def draw(self, can, x, y):
        # move to the bottom-left corner
        midx = x
        midy = y
        d_len = self.size / 2.0
        r_len = self.size * 1.414 / 2.0
        can.line(self.line_style, x - d_len, y - d_len, x + d_len, y + d_len)
        can.line(self.line_style, x + d_len, y - d_len, x - d_len, y + d_len)
        can.line(self.line_style, midx, y - r_len, midx, y + r_len)
        can.line(self.line_style, x - r_len, midy, x + r_len, midy)


class Null(T):
    """This tickmark doesn't draw anything. All the attributes are ignored."""

    def __init__(self):
        self.line_style = None
        self.fill_style = None
        self.size = -1

    def draw(self, can, x, y):
        pass


standards = object_set.T()


def _intern(style):
    standards.add(style)
    return style


square = _intern(Square())
square3 = _intern(Square(size=3))
square5 = square
x = _intern(X())
x3 = _intern(X(size=3))
x5 = x
star = _intern(Star())
star3 = _intern(Star(size=3))
star5 = star
plus = _intern(Plus())
plus3 = _intern(Plus(size=3))
plus5 = plus
dia = _intern(Diamond())
dia3 = _intern(Diamond(size=3))
dia5 = dia
tri = _intern(Triangle())
tri3 = _intern(Triangle(size=3))
tri5 = tri
dtri = _intern(DownTriangle())
dtri3 = _intern(DownTriangle(size=3))
dtri5 = dtri
circle1 = _intern(Circle(size=1))
circle2 = _intern(Circle(size=3))
circle3 = _intern(Circle(size=5))
blacksquare = _intern(Square(fill_style=fill_style.black))
blacksquare3 = _intern(Square(size=3, fill_style=fill_style.black))
blackdia = _intern(Diamond(fill_style=fill_style.black))
blackdia3 = _intern(Diamond(size=3, fill_style=fill_style.black))
blacktri = _intern(Triangle(fill_style=fill_style.black))
blacktri3 = _intern(Triangle(size=3, fill_style=fill_style.black))
blackdtri = _intern(DownTriangle(fill_style=fill_style.black))
blackdtri3 = _intern(DownTriangle(size=3, fill_style=fill_style.black))
blackcircle1 = _intern(Circle(size=1, fill_style=fill_style.black))
blackcircle3 = _intern(Circle(size=3, fill_style=fill_style.black))
gray70square = _intern(Square(fill_style=fill_style.gray70))
gray70square3 = _intern(Square(size=3, fill_style=fill_style.gray70))
gray70dia = _intern(Diamond(fill_style=fill_style.gray70))
gray70dia3 = _intern(Diamond(size=3, fill_style=fill_style.gray70))
gray70tri = _intern(Triangle(fill_style=fill_style.gray70))
gray70tri3 = _intern(Triangle(size=3, fill_style=fill_style.gray70))
gray70dtri = _intern(DownTriangle(fill_style=fill_style.gray70))
gray70dtri3 = _intern(DownTriangle(size=3, fill_style=fill_style.gray70))
gray70circle1 = _intern(Circle(size=1, fill_style=fill_style.gray70))
gray70circle3 = _intern(Circle(size=3, fill_style=fill_style.gray70))
default = _intern(Null())



#
# Copyright (C) 2000-2005 by Yasushi Saito (yasushi.saito@gmail.com)
# 
# Jockey is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
#
# Jockey is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
from . import chart_object
from . import fill_style
from . import legend
from . import line_style
from . import range_plot_doc
from . import pychart_util
from .pychart_types import *
from .scaling import *


class T(chart_object.T):
    __doc__ = range_plot_doc.doc
    keys = {
        "data": (AnyType, None, pychart_util.data_desc),
        "label": (StringType, "???", pychart_util.label_desc),
        "xcol": (IntType, 0, pychart_util.xcol_desc),
        "min_col": (IntType, 1,
                    "The lower bound of the sweep is extracted from "
                    + "this column of data."),
        "max_col": (IntType, 2,
                    "The upper bound of the sweep is extracted from "
                    + "this column of data."),
        "line_style": (line_style.T, line_style.default,
                       "The style of the boundary line."),
        "fill_style": (fill_style.T, fill_style.default,
                       ""),
    }

    ##AUTOMATICALLY GENERATED

    ##END AUTOMATICALLY GENERATED

    def check_integrity(self):
        chart_object.T.check_integrity(self)

    def get_data_range(self, which):
        if which == 'X':
            return pychart_util.get_data_range(self.data, self.xcol)
        else:
            ymax = (pychart_util.get_data_range(self.data, self.max_col))[1]
            ymin = (pychart_util.get_data_range(self.data, self.min_col))[0]
            return (ymin, ymax)

    def get_legend_entry(self):
        if self.label:
            return legend.Entry(line_style=self.line_style,
                                fill_style=self.fill_style,
                                label=self.label)
        return None

    def draw(self, ar, can):

        prevPair = None

        xmin = 999999
        xmax = -999999
        ymin = 999999
        ymax = -999999

        # Draw the boundary in a single stroke.
        can.gsave()
        can.newpath()
        for pair in self.data:
            x = pair[self.xcol]
            y = pychart_util.get_sample_val(pair, self.max_col)
            if y == None:
                continue

            xmin = min(xmin, ar.x_pos(x))
            xmax = max(xmax, ar.x_pos(x))
            ymin = min(ymin, ar.y_pos(y))
            ymax = max(ymax, ar.y_pos(y))
            if prevPair != None:
                can.lineto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
            else:
                can.moveto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
            prevPair = pair

        for i in range(len(self.data) - 1, -1, -1):
            pair = self.data[i]
            x = pair[self.xcol]
            y = pychart_util.get_sample_val(pair, self.min_col)
            if None in (x, y):
                continue

            xmin = min(xmin, ar.x_pos(x))
            xmax = max(xmax, ar.x_pos(x))
            ymin = min(ymin, ar.y_pos(y))
            ymax = max(ymax, ar.y_pos(y))
            can.lineto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
        can.closepath()

        # create a clip region, and fill it.
        can.clip_sub()
        can.fill_with_pattern(self.fill_style, xmin, ymin, xmax, ymax)
        can.grestore()

        if self.line_style:
            # draw the boundary.
            prevPair = None
            can.newpath()
            can.set_line_style(self.line_style)
            for pair in self.data:
                x = pair[self.xcol]
                y = pychart_util.get_sample_val(pair, self.min_col)
                if None in (x, y):
                    continue

                if prevPair != None:
                    can.lineto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
                else:
                    can.moveto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
                prevPair = pair
            can.stroke()

            prevPair = None
            can.newpath()
            can.set_line_style(self.line_style)
            for pair in self.data:
                x = pair[self.xcol]
                y = pychart_util.get_sample_val(pair, self.max_col)
                if y == None:
                    continue

                if prevPair != None:
                    can.lineto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
                else:
                    can.moveto(xscale(ar.x_pos(x)), yscale(ar.y_pos(y)))
                prevPair = pair
            can.stroke()
